@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
 * This file is part of 0 A.D.
 *
 * 0 A.D. is free software: you can redistribute it and/or modify
@@ -21,21 +21,22 @@
 
 #include "graphics/HFTracer.h"
 #include "graphics/Terrain.h"
-#include "graphics/scripting/JSInterface_GameView.h"
+#include "i18n/L10n.h"
 #include "lib/input.h"
 #include "lib/timer.h"
 #include "maths/MathUtil.h"
 #include "maths/Matrix3D.h"
 #include "maths/Quaternion.h"
+#include "ps/CLogger.h"
 #include "ps/ConfigDB.h"
 #include "ps/Game.h"
 #include "ps/Globals.h"
 #include "ps/Hotkey.h"
-#include "ps/Joystick.h"
 #include "ps/Pyrogenesis.h"
 #include "ps/TouchInput.h"
 #include "ps/World.h"
 #include "renderer/Renderer.h"
+#include "renderer/SceneRenderer.h"
 #include "renderer/WaterManager.h"
 #include "simulation2/Simulation2.h"
 #include "simulation2/components/ICmpPosition.h"
@@ -57,6 +58,7 @@ CCameraController::CCameraController(CCamera& camera)
 	  // Dummy values (these will be filled in by the config file)
 	  m_ViewScrollSpeed(0),
 	  m_ViewScrollSpeedModifier(1),
+	  m_ViewScrollMouseDetectDistance(3),
 	  m_ViewRotateXSpeed(0),
 	  m_ViewRotateXMin(0),
 	  m_ViewRotateXMax(0),
@@ -75,12 +77,6 @@ CCameraController::CCameraController(CCamera& camera)
 	  m_ViewFOV(DEGTORAD(45.f)),
 	  m_ViewNear(2.f),
 	  m_ViewFar(4096.f),
-	  m_JoystickPanX(-1),
-	  m_JoystickPanY(-1),
-	  m_JoystickRotateX(-1),
-	  m_JoystickRotateY(-1),
-	  m_JoystickZoomIn(-1),
-	  m_JoystickZoomOut(-1),
 	  m_HeightSmoothness(0.5f),
 	  m_HeightMin(16.f),
 
@@ -109,6 +105,7 @@ void CCameraController::LoadConfig()
 {
 	CFG_GET_VAL("view.scroll.speed", m_ViewScrollSpeed);
 	CFG_GET_VAL("view.scroll.speed.modifier", m_ViewScrollSpeedModifier);
+	CFG_GET_VAL("view.scroll.mouse.detectdistance", m_ViewScrollMouseDetectDistance);
 	CFG_GET_VAL("view.rotate.x.speed", m_ViewRotateXSpeed);
 	CFG_GET_VAL("view.rotate.x.min", m_ViewRotateXMin);
 	CFG_GET_VAL("view.rotate.x.max", m_ViewRotateXMax);
@@ -125,13 +122,6 @@ void CCameraController::LoadConfig()
 	CFG_GET_VAL("view.zoom.default", m_ViewZoomDefault);
 	CFG_GET_VAL("view.zoom.speed.modifier", m_ViewZoomSpeedModifier);
 
-	CFG_GET_VAL("joystick.camera.pan.x", m_JoystickPanX);
-	CFG_GET_VAL("joystick.camera.pan.y", m_JoystickPanY);
-	CFG_GET_VAL("joystick.camera.rotate.x", m_JoystickRotateX);
-	CFG_GET_VAL("joystick.camera.rotate.y", m_JoystickRotateY);
-	CFG_GET_VAL("joystick.camera.zoom.in", m_JoystickZoomIn);
-	CFG_GET_VAL("joystick.camera.zoom.out", m_JoystickZoomOut);
-
 	CFG_GET_VAL("view.height.smoothness", m_HeightSmoothness);
 	CFG_GET_VAL("view.height.min", m_HeightMin);
 
@@ -194,16 +184,16 @@ void CCameraController::Update(const float deltaRealTime)
 		moveForward += m_ViewDragSpeed * -mouse_dy;
 	}
 
-	if (g_mouse_active)
+	if (g_mouse_active && m_ViewScrollMouseDetectDistance > 0)
 	{
-		if (g_mouse_x >= g_xres - 2 && g_mouse_x < g_xres)
+		if (g_mouse_x >= g_xres - m_ViewScrollMouseDetectDistance && g_mouse_x < g_xres)
 			moveRightward += m_ViewScrollSpeed * deltaRealTime;
-		else if (g_mouse_x <= 3 && g_mouse_x >= 0)
+		else if (g_mouse_x < m_ViewScrollMouseDetectDistance && g_mouse_x >= 0)
 			moveRightward -= m_ViewScrollSpeed * deltaRealTime;
 
-		if (g_mouse_y >= g_yres - 2 && g_mouse_y < g_yres)
+		if (g_mouse_y >= g_yres - m_ViewScrollMouseDetectDistance && g_mouse_y < g_yres)
 			moveForward -= m_ViewScrollSpeed * deltaRealTime;
-		else if (g_mouse_y <= 3 && g_mouse_y >= 0)
+		else if (g_mouse_y < m_ViewScrollMouseDetectDistance && g_mouse_y >= 0)
 			moveForward += m_ViewScrollSpeed * deltaRealTime;
 	}
 
@@ -216,22 +206,6 @@ void CCameraController::Update(const float deltaRealTime)
 	if (HotkeyIsPressed("camera.down"))
 		moveForward -= m_ViewScrollSpeed * deltaRealTime;
 
-	if (g_Joystick.IsEnabled())
-	{
-		// This could all be improved with extra speed and sensitivity settings
-		// (maybe use pow to allow finer control?), and inversion settings
-
-		moveRightward += g_Joystick.GetAxisValue(m_JoystickPanX) * m_ViewScrollSpeed * deltaRealTime;
-		moveForward -= g_Joystick.GetAxisValue(m_JoystickPanY) * m_ViewScrollSpeed * deltaRealTime;
-
-		m_RotateX.AddSmoothly(g_Joystick.GetAxisValue(m_JoystickRotateX) * m_ViewRotateXSpeed * deltaRealTime);
-		m_RotateY.AddSmoothly(-g_Joystick.GetAxisValue(m_JoystickRotateY) * m_ViewRotateYSpeed * deltaRealTime);
-
-		// Use a +1 bias for zoom because I want this to work with trigger buttons that default to -1
-		m_Zoom.AddSmoothly((g_Joystick.GetAxisValue(m_JoystickZoomIn) + 1.0f) / 2.0f * m_ViewZoomSpeed * deltaRealTime);
-		m_Zoom.AddSmoothly(-(g_Joystick.GetAxisValue(m_JoystickZoomOut) + 1.0f) / 2.0f * m_ViewZoomSpeed * deltaRealTime);
-	}
-
 	if (moveRightward || moveForward)
 	{
 		// Break out of following mode when the user starts scrolling
@@ -607,7 +581,7 @@ void CCameraController::FocusHeight(bool smooth)
 
 	const float ground = std::max(
 		g_Game->GetWorld()->GetTerrain()->GetExactGroundLevel(nearPoint.X, nearPoint.Z),
-		g_Renderer.GetWaterManager()->m_WaterHeight);
+		g_Renderer.GetSceneRenderer().GetWaterManager().m_WaterHeight);
 
 	// filter ground levels for smooth camera movement
 	const float filtered_near_ground = g_Game->GetWorld()->GetTerrain()->GetFilteredGroundLevel(nearPoint.X, nearPoint.Z, near_radius);
@@ -616,7 +590,7 @@ void CCameraController::FocusHeight(bool smooth)
 	// filtered maximum visible ground level in view
 	const float filtered_ground = std::max(
 		std::max(filtered_near_ground, filtered_pivot_ground),
-		g_Renderer.GetWaterManager()->m_WaterHeight);
+		g_Renderer.GetSceneRenderer().GetWaterManager().m_WaterHeight);
 
 	// target camera height above pivot point
 	const float pivot_height = -forwards.Y * (m_Zoom.GetSmoothedValue() - m_ViewNear);
@@ -682,33 +656,39 @@ InReaction CCameraController::HandleEvent(const SDL_Event_* ev)
 		else if (hotkey == "camera.scroll.speed.increase")
 		{
 			m_ViewScrollSpeed *= m_ViewScrollSpeedModifier;
+			LOGMESSAGERENDER(g_L10n.Translate("Scroll speed increased to %.1f"), m_ViewScrollSpeed);
 			return IN_HANDLED;
 		}
 		else if (hotkey == "camera.scroll.speed.decrease")
 		{
 			m_ViewScrollSpeed /= m_ViewScrollSpeedModifier;
+			LOGMESSAGERENDER(g_L10n.Translate("Scroll speed decreased to %.1f"), m_ViewScrollSpeed);
 			return IN_HANDLED;
 		}
 		else if (hotkey == "camera.rotate.speed.increase")
 		{
 			m_ViewRotateXSpeed *= m_ViewRotateSpeedModifier;
 			m_ViewRotateYSpeed *= m_ViewRotateSpeedModifier;
+			LOGMESSAGERENDER(g_L10n.Translate("Rotate speed increased to X=%.3f, Y=%.3f"), m_ViewRotateXSpeed, m_ViewRotateYSpeed);
 			return IN_HANDLED;
 		}
 		else if (hotkey == "camera.rotate.speed.decrease")
 		{
 			m_ViewRotateXSpeed /= m_ViewRotateSpeedModifier;
 			m_ViewRotateYSpeed /= m_ViewRotateSpeedModifier;
+			LOGMESSAGERENDER(g_L10n.Translate("Rotate speed decreased to X=%.3f, Y=%.3f"), m_ViewRotateXSpeed, m_ViewRotateYSpeed);
 			return IN_HANDLED;
 		}
 		else if (hotkey == "camera.zoom.speed.increase")
 		{
 			m_ViewZoomSpeed *= m_ViewZoomSpeedModifier;
+			LOGMESSAGERENDER(g_L10n.Translate("Zoom speed increased to %.1f"), m_ViewZoomSpeed);
 			return IN_HANDLED;
 		}
 		else if (hotkey == "camera.zoom.speed.decrease")
 		{
 			m_ViewZoomSpeed /= m_ViewZoomSpeedModifier;
+			LOGMESSAGERENDER(g_L10n.Translate("Zoom speed decreased to %.1f"), m_ViewZoomSpeed);
 			return IN_HANDLED;
 		}
 		return IN_PASS;
