@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -15,40 +15,33 @@
  * along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/*
- * Mesh object with texture and skinning information
- */
-
 #include "precompiled.h"
 
 #include "Model.h"
 
-#include "Decal.h"
-#include "ModelDef.h"
-#include "maths/Quaternion.h"
+#include "graphics/Decal.h"
+#include "graphics/MeshManager.h"
+#include "graphics/ModelDef.h"
+#include "graphics/ObjectEntry.h"
+#include "graphics/SkeletonAnim.h"
+#include "graphics/SkeletonAnimDef.h"
 #include "maths/BoundingBoxAligned.h"
-#include "SkeletonAnim.h"
-#include "SkeletonAnimDef.h"
-#include "SkeletonAnimManager.h"
-#include "MeshManager.h"
-#include "ObjectEntry.h"
-#include "lib/res/graphics/ogl_tex.h"
-#include "lib/res/h_mgr.h"
+#include "maths/Quaternion.h"
 #include "lib/sysdep/rtl.h"
-#include "ps/Profile.h"
 #include "ps/CLogger.h"
+#include "ps/CStrInternStatic.h"
+#include "ps/Profile.h"
 #include "renderer/RenderingOptions.h"
-#include "simulation2/Simulation2.h"
 #include "simulation2/components/ICmpTerrain.h"
 #include "simulation2/components/ICmpWaterManager.h"
+#include "simulation2/Simulation2.h"
 
 
 /////////////////////////////////////////////////////////////////////////////////////////////////////////////
 // Constructor
-CModel::CModel(CSkeletonAnimManager& skeletonAnimManager, CSimulation2& simulation)
+CModel::CModel(CSimulation2& simulation)
 	: m_Flags(0), m_Anim(NULL), m_AnimTime(0), m_Simulation(simulation),
-	m_BoneMatrices(NULL), m_AmmoPropPoint(NULL), m_AmmoLoadedProp(0),
-	m_SkeletonAnimManager(skeletonAnimManager)
+	m_BoneMatrices(NULL), m_AmmoPropPoint(NULL), m_AmmoLoadedProp(0)
 {
 }
 
@@ -130,66 +123,16 @@ void CModel::CalcBounds()
 // CalcObjectBounds: calculate object space bounds of this model, based solely on vertex positions
 void CModel::CalcStaticObjectBounds()
 {
-	m_ObjectBounds.SetEmpty();
-
-	size_t numverts=m_pModelDef->GetNumVertices();
-	SModelVertex* verts=m_pModelDef->GetVertices();
-
-	for (size_t i=0;i<numverts;i++) {
-		m_ObjectBounds+=verts[i].m_Coords;
-	}
+	PROFILE2("CalcStaticObjectBounds");
+	m_pModelDef->GetMaxBounds(nullptr, !(m_Flags & MODELFLAG_NOLOOPANIMATION), m_ObjectBounds);
 }
 
 /////////////////////////////////////////////////////////////////////////////////////////////////////////////
 // CalcAnimatedObjectBound: calculate bounds encompassing all vertex positions for given animation
 void CModel::CalcAnimatedObjectBounds(CSkeletonAnimDef* anim, CBoundingBoxAligned& result)
 {
-	result.SetEmpty();
-
-	// Set the current animation on which to perform calculations (if it's necessary)
-	if (anim != m_Anim->m_AnimDef)
-	{
-		CSkeletonAnim dummyanim;
-		dummyanim.m_AnimDef=anim;
-		if (!SetAnimation(&dummyanim)) return;
-	}
-
-	size_t numverts=m_pModelDef->GetNumVertices();
-	SModelVertex* verts=m_pModelDef->GetVertices();
-
-	// Remove any transformations, so that we calculate the bounding box
-	// at the origin. The box is later re-transformed onto the object, without
-	// having to recalculate the size of the box.
-	CMatrix3D transform, oldtransform = GetTransform();
-	CModelAbstract* oldparent = m_Parent;
-
-	m_Parent = 0;
-	transform.SetIdentity();
-	CRenderableObject::SetTransform(transform);
-
-	// Following seems to stomp over the current animation time - which, unsurprisingly,
-	// introduces artefacts in the currently playing animation. Save it here and restore it
-	// at the end.
-	float AnimTime = m_AnimTime;
-
-	// iterate through every frame of the animation
-	for (size_t j=0;j<anim->GetNumFrames();j++) {
-		m_PositionValid = false;
-		ValidatePosition();
-
-		// extend bounds by vertex positions at the frame
-		for (size_t i=0;i<numverts;i++)
-		{
-			result += CModelDef::SkinPoint(verts[i], GetAnimatedBoneMatrices());
-		}
-		// advance to next frame
-		m_AnimTime += anim->GetFrameTime();
-	}
-
-	m_PositionValid = false;
-	m_Parent = oldparent;
-	SetTransform(oldtransform);
-	m_AnimTime = AnimTime;
+	PROFILE2("CalcAnimatedObjectBounds");
+	m_pModelDef->GetMaxBounds(anim, !(m_Flags & MODELFLAG_NOLOOPANIMATION), result);
 }
 
 /////////////////////////////////////////////////////////////////////////////////////////////////////////////
@@ -253,43 +196,6 @@ const CBoundingBoxAligned CModel::GetObjectSelectionBoundsRec()
 	return objBounds;
 }
 
-/////////////////////////////////////////////////////////////////////////////////////////////////////////////
-// BuildAnimation: load raw animation frame animation from given file, and build a
-// animation specific to this model
-CSkeletonAnim* CModel::BuildAnimation(const VfsPath& pathname, const CStr& name, const CStr& ID, int frequency, float speed, float actionpos, float actionpos2, float soundpos)
-{
-	CSkeletonAnimDef* def = m_SkeletonAnimManager.GetAnimation(pathname);
-	if (!def)
-		return NULL;
-
-	CSkeletonAnim* anim = new CSkeletonAnim();
-	anim->m_Name = name;
-	anim->m_ID = ID;
-	anim->m_Frequency = frequency;
-	anim->m_AnimDef = def;
-	anim->m_Speed = speed;
-
-	if (actionpos == -1.f)
-		anim->m_ActionPos = -1.f;
-	else
-		anim->m_ActionPos = actionpos * anim->m_AnimDef->GetDuration();
-
-	if (actionpos2 == -1.f)
-		anim->m_ActionPos2 = -1.f;
-	else
-		anim->m_ActionPos2 = actionpos2 * anim->m_AnimDef->GetDuration();
-
-	if (soundpos == -1.f)
-		anim->m_SoundPos = -1.f;
-	else
-		anim->m_SoundPos = soundpos * anim->m_AnimDef->GetDuration();
-
-	anim->m_ObjectBounds.SetEmpty();
-	InvalidateBounds();
-
-	return anim;
-}
-
 /////////////////////////////////////////////////////////////////////////////////////////////////////////////
 // Update: update this model to the given time, in msec
 void CModel::UpdateTo(float time)
@@ -499,10 +405,6 @@ void CModel::CopyAnimationFrom(CModel* source)
 	m_Anim = source->m_Anim;
 	m_AnimTime = source->m_AnimTime;
 
-	m_Flags &= ~MODELFLAG_CASTSHADOWS;
-	if (source->m_Flags & MODELFLAG_CASTSHADOWS)
-		m_Flags |= MODELFLAG_CASTSHADOWS;
-
 	m_ObjectBounds.SetEmpty();
 	InvalidateBounds();
 }
@@ -592,7 +494,7 @@ CModelAbstract* CModel::FindFirstAmmoProp()
 // Clone: return a clone of this model
 CModelAbstract* CModel::Clone() const
 {
-	CModel* clone = new CModel(m_SkeletonAnimManager, m_Simulation);
+	CModel* clone = new CModel(m_Simulation);
 	clone->m_ObjectBounds = m_ObjectBounds;
 	clone->InitModel(m_pModelDef);
 	clone->SetMaterial(m_Material);
@@ -628,10 +530,7 @@ void CModel::AddFlagsRec(int flags)
 	m_Flags |= flags;
 
 	if (flags & MODELFLAG_IGNORE_LOS)
-	{
 		m_Material.AddShaderDefine(str_IGNORE_LOS, str_1);
-		m_Material.RecomputeCombinedShaderDefines();
-	}
 
 	for (size_t i = 0; i < m_Props.size(); ++i)
 		if (m_Props[i].m_Model->ToCModel())
@@ -643,7 +542,6 @@ void CModel::RemoveShadowsRec()
 	m_Flags &= ~MODELFLAG_CASTSHADOWS;
 
 	m_Material.AddShaderDefine(str_DISABLE_RECEIVE_SHADOWS, str_1);
-	m_Material.RecomputeCombinedShaderDefines();
 
 	for (size_t i = 0; i < m_Props.size(); ++i)
 	{
