@@ -1,4 +1,4 @@
-/* Copyright (C) 2021 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -15,22 +15,61 @@
  * along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/*
- * Defines a raw 3d model.
- */
-
 #include "precompiled.h"
 
 #include "ModelDef.h"
+
 #include "graphics/SkeletonAnimDef.h"
 #include "lib/sysdep/arch/x86_x64/simd.h"
-#include "ps/FileIo.h"
 #include "maths/Vector4D.h"
+#include "ps/FileIo.h"
 
 #if COMPILER_HAS_SSE
 # include <xmmintrin.h>
 #endif
 
+void CModelDef::GetMaxBounds(CSkeletonAnimDef* anim, bool loop, CBoundingBoxAligned& result)
+{
+	const u32 animIndex = anim ? anim->m_UID : 0;
+
+	std::unordered_map<u32, CBoundingBoxAligned>::const_iterator it = m_MaxBoundsPerAnimDef.find(animIndex);
+	if (it != m_MaxBoundsPerAnimDef.end())
+	{
+		result = it->second;
+		return;
+	}
+
+	size_t numverts = GetNumVertices();
+	SModelVertex* verts = GetVertices();
+
+	if (!anim)
+	{
+		for (size_t i = 0; i < numverts; ++i)
+			result += verts[i].m_Coords;
+		m_MaxBoundsPerAnimDef[animIndex] = result;
+		return;
+	}
+	ENSURE(animIndex != 0);
+	CMatrix3D* inverseBindBoneMatrix = GetInverseBindBoneMatrices();
+	std::vector<CMatrix3D> boneMatrix(anim->GetNumKeys());
+
+	const size_t numFrames = anim->GetNumFrames();
+	const float frameTime = anim->GetFrameTime();
+	const size_t numBones = GetNumBones();
+
+	// NB: by using frames, the bounds are technically pessimistic,
+	// since interpolation between frames can put vertices farther.
+	for (size_t j = 0; j < numFrames; ++j)
+	{
+		anim->BuildBoneMatrices(j * frameTime, boneMatrix.data(), loop);
+		for (size_t i = 0; i < numBones; ++i)
+			boneMatrix[i] = boneMatrix[i] * inverseBindBoneMatrix[i];
+		for (size_t i = 0; i < numverts; ++i)
+			result += SkinPoint(verts[i], boneMatrix.data());
+	}
+	m_MaxBoundsPerAnimDef[animIndex] = result;
+}
+
 CVector3D CModelDef::SkinPoint(const SModelVertex& vtx,
 							   const CMatrix3D newPoseMatrices[])
 {
@@ -161,10 +200,11 @@ static void SkinPointsAndNormalsSSE(
 		const CMatrix3D& mtx = newPoseMatrices[blendIndices[j]];
 
 		// Loads matrix to xmm registers.
-		col0 = _mm_load_ps(mtx._data);
-		col1 = _mm_load_ps(mtx._data + 4);
-		col2 = _mm_load_ps(mtx._data + 8);
-		col3 = _mm_load_ps(mtx._data + 12);
+		const float* data = mtx.AsFloatArray().data();
+		col0 = _mm_load_ps(data);
+		col1 = _mm_load_ps(data + 4);
+		col2 = _mm_load_ps(data + 8);
+		col3 = _mm_load_ps(data + 12);
 
 		// Loads and computes vertex coordinates.
 		vec0 = _mm_load1_ps(&vtx.m_Coords.X);	// v0 = [x, x, x, x]
@@ -296,7 +336,8 @@ CModelDef* CModelDef::Load(const VfsPath& filename, const VfsPath& name)
 		mdef->m_NumUVsPerVertex = unpacker.UnpackSize();
 	}
 
-	mdef->m_pVertices=new SModelVertex[mdef->m_NumVertices];
+	mdef->m_pVertices = new SModelVertex[mdef->m_NumVertices];
+	mdef->m_UVCoordinates.reserve(mdef->m_NumVertices * mdef->m_NumUVsPerVertex);
 
 	for (size_t i = 0; i < mdef->m_NumVertices; ++i)
 	{
@@ -307,8 +348,7 @@ CModelDef* CModelDef::Load(const VfsPath& filename, const VfsPath& name)
 		{
 			float uv[2];
 			unpacker.UnpackRaw(&uv[0], 8);
-			mdef->m_pVertices[i].m_UVs.push_back(uv[0]);
-			mdef->m_pVertices[i].m_UVs.push_back(uv[1]);
+			mdef->m_UVCoordinates.emplace_back(uv[0], uv[1]);
 		}
 
 		unpacker.UnpackRaw(&mdef->m_pVertices[i].m_Blend, sizeof(SVertexBlend));
