@@ -1,4 +1,4 @@
-/* Copyright (C) 2019 Wildfire Games.
+/* Copyright (C) 2021 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -23,18 +23,15 @@
 
 CImage::CImage(CGUI& pGUI)
 	: IGUIObject(pGUI),
-	  m_Sprite(),
-	  m_CellID()
+	  m_Sprite(this, "sprite")
 {
-	RegisterSetting("sprite", m_Sprite);
-	RegisterSetting("cell_id", m_CellID);
 }
 
 CImage::~CImage()
 {
 }
 
-void CImage::Draw()
+void CImage::Draw(CCanvas2D& canvas)
 {
-	m_pGUI.DrawSprite(m_Sprite, m_CellID, GetBufferedZ(), m_CachedActualSize);
+	m_pGUI.DrawSprite(m_Sprite, canvas, m_CachedActualSize);
 }
