@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2021 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -79,6 +79,12 @@ General design and rules:
 
 #include <cstring>
 
+// Gloox leaves some #define up, we need to undefine them.
+#undef lookup
+#undef lookup2
+#undef deflookup
+#undef deflookup2
+
 #if OS_WIN
 #define GLOOXWRAPPER_API __declspec(dllexport)
 #else
@@ -408,6 +414,7 @@ namespace glooxwrapper
 		bool connect(bool block = true);
 		gloox::ConnectionError recv(int timeout = -1);
 		const string getID() const;
+		const string getJID() const;
 		void send(const IQ& iq);
 
 		void setTls(gloox::TLSPolicy tls);
@@ -558,7 +565,7 @@ namespace glooxwrapper
 		NONCOPYABLE(Registration);
 	private:
 		gloox::Registration* m_Wrapped;
-		std::list<shared_ptr<gloox::RegistrationHandler> > m_RegistrationHandlers;
+		std::list<std::shared_ptr<gloox::RegistrationHandler> > m_RegistrationHandlers;
 	public:
 		Registration(Client* parent);
 		~Registration();
@@ -666,7 +673,7 @@ namespace glooxwrapper
 			Session(gloox::Jingle::Session* wrapped, bool owned);
 			~Session();
 
-			bool sessionInitiate(char* ipStr, uint16_t port);
+			bool sessionInitiate(const char* ipStr, uint16_t port);
 		};
 
 		class GLOOXWRAPPER_API SessionHandler
