@@ -1,4 +1,4 @@
-/* Copyright (C) 2021 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -33,13 +33,23 @@
 #include "ps/Compress.h"
 #include "ps/CStr.h"
 #include "ps/Game.h"
+#include "ps/Hashing.h"
 #include "ps/Loader.h"
 #include "ps/Profile.h"
 #include "ps/Threading.h"
 #include "scriptinterface/ScriptInterface.h"
+#include "scriptinterface/JSON.h"
 #include "simulation2/Simulation2.h"
 #include "network/StunClient.h"
 
+/**
+ * Once ping goes above turn length * command delay,
+ * the game will start 'freezing' for other clients while we catch up.
+ * Since commands are sent client -> server -> client, divide by 2.
+ * (duplicated in NetServer.cpp to avoid having to fetch the constants in a header file)
+ */
+constexpr u32 NETWORK_BAD_PING = DEFAULT_TURN_LENGTH * COMMAND_DELAY_MP / 2;
+
 CNetClient *g_NetClient = NULL;
 
 /**
@@ -50,8 +60,8 @@ class CNetFileReceiveTask_ClientRejoin : public CNetFileReceiveTask
 {
 	NONCOPYABLE(CNetFileReceiveTask_ClientRejoin);
 public:
-	CNetFileReceiveTask_ClientRejoin(CNetClient& client)
-		: m_Client(client)
+	CNetFileReceiveTask_ClientRejoin(CNetClient& client, const CStr& initAttribs)
+		: m_Client(client), m_InitAttributes(initAttribs)
 	{
 	}
 
@@ -64,19 +74,19 @@ public:
 
 		// Pretend the server told us to start the game
 		CGameStartMessage start;
+		start.m_InitAttributes = m_InitAttributes;
 		m_Client.HandleMessage(&start);
 	}
 
 private:
 	CNetClient& m_Client;
+	CStr m_InitAttributes;
 };
 
-CNetClient::CNetClient(CGame* game, bool isLocalClient) :
+CNetClient::CNetClient(CGame* game) :
 	m_Session(NULL),
 	m_UserName(L"anonymous"),
 	m_HostID((u32)-1), m_ClientTurnManager(NULL), m_Game(game),
-	m_GameAttributes(game->GetSimulation2()->GetScriptInterface().GetGeneralJSContext()),
-	m_IsLocalClient(isLocalClient),
 	m_LastConnectionCheck(0),
 	m_ServerAddress(),
 	m_ServerPort(0),
@@ -96,9 +106,7 @@ CNetClient::CNetClient(CGame* game, bool isLocalClient) :
 	AddTransition(NCS_HANDSHAKE, (uint)NMT_SERVER_HANDSHAKE_RESPONSE, NCS_AUTHENTICATE, (void*)&OnHandshakeResponse, context);
 
 	AddTransition(NCS_AUTHENTICATE, (uint)NMT_AUTHENTICATE, NCS_AUTHENTICATE, (void*)&OnAuthenticateRequest, context);
-	AddTransition(NCS_AUTHENTICATE, (uint)NMT_AUTHENTICATE_RESULT, NCS_INITIAL_GAMESETUP, (void*)&OnAuthenticate, context);
-
-	AddTransition(NCS_INITIAL_GAMESETUP, (uint)NMT_GAME_SETUP, NCS_PREGAME, (void*)&OnGameSetup, context);
+	AddTransition(NCS_AUTHENTICATE, (uint)NMT_AUTHENTICATE_RESULT, NCS_PREGAME, (void*)&OnAuthenticate, context);
 
 	AddTransition(NCS_PREGAME, (uint)NMT_CHAT, NCS_PREGAME, (void*)&OnChat, context);
 	AddTransition(NCS_PREGAME, (uint)NMT_READY, NCS_PREGAME, (void*)&OnReady, context);
@@ -170,25 +178,39 @@ void CNetClient::SetUserName(const CStrW& username)
 	m_UserName = username;
 }
 
-void CNetClient::SetHostingPlayerName(const CStr& hostingPlayerName)
+void CNetClient::SetHostJID(const CStr& jid)
 {
-	m_HostingPlayerName = hostingPlayerName;
+	m_HostJID = jid;
 }
 
 void CNetClient::SetGamePassword(const CStr& hashedPassword)
 {
-	m_Password = hashedPassword;
+	// Hash on top with the user's name, to make sure not all
+	// hashing data is in control of the host.
+	m_Password = HashCryptographically(hashedPassword, m_UserName.ToUTF8());
 }
 
+void CNetClient::SetControllerSecret(const std::string& secret)
+{
+	m_ControllerSecret = secret;
+}
+
+
 bool CNetClient::SetupConnection(ENetHost* enetClient)
 {
 	CNetClientSession* session = new CNetClientSession(*this);
-	bool ok = session->Connect(m_ServerAddress, m_ServerPort, m_IsLocalClient, enetClient);
+	bool ok = session->Connect(m_ServerAddress, m_ServerPort, enetClient);
 	SetAndOwnSession(session);
-	m_PollingThread = std::thread(Threading::HandleExceptions<CNetClientSession::RunNetLoop>::Wrapper, m_Session);
+	if (ok)
+		m_PollingThread = std::thread(Threading::HandleExceptions<CNetClientSession::RunNetLoop>::Wrapper, m_Session);
 	return ok;
 }
 
+void CNetClient::SetupConnectionViaLobby()
+{
+	g_XmppClient->SendIqGetConnectionData(m_HostJID, m_Password, m_UserName.ToUTF8(), false);
+}
+
 void CNetClient::SetupServerData(CStr address, u16 port, bool stun)
 {
 	ENSURE(!m_Session);
@@ -210,7 +232,7 @@ void CNetClient::HandleGetServerDataFailed(const CStr& error)
 	);
 }
 
-bool CNetClient::TryToConnect(const CStr& hostJID)
+bool CNetClient::TryToConnect(const CStr& hostJID, bool localNetwork)
 {
 	if (m_Session)
 		return false;
@@ -224,30 +246,23 @@ bool CNetClient::TryToConnect(const CStr& hostJID)
 		return false;
 	}
 
-	ENetHost* enetClient = nullptr;
+	ENetAddress hostAddr{ ENET_HOST_ANY, ENET_PORT_ANY };
+	ENetHost* enetClient = enet_host_create(&hostAddr, 1, 1, 0, 0);
+
+	if (!enetClient)
+	{
+		PushGuiMessage(
+			"type", "netstatus",
+			"status", "disconnected",
+			"reason", static_cast<i32>(NDR_STUN_PORT_FAILED));
+		return false;
+	}
+
+	CStr ip;
+	u16 port = 0;
 	if (g_XmppClient && m_UseSTUN)
 	{
-		// Find an unused port
-		for (int i = 0; i < 5 && !enetClient; ++i)
-		{
-			// Ports below 1024 are privileged on unix
-			u16 port = 1024 + rand() % (UINT16_MAX - 1024);
-			ENetAddress hostAddr{ ENET_HOST_ANY, port };
-			enetClient = enet_host_create(&hostAddr, 1, 1, 0, 0);
-			++hostAddr.port;
-		}
-
-		if (!enetClient)
-		{
-			PushGuiMessage(
-				"type", "netstatus",
-				"status", "disconnected",
-				"reason", static_cast<i32>(NDR_STUN_PORT_FAILED));
-			return false;
-		}
-
-		StunClient::StunEndpoint stunEndpoint;
-		if (!StunClient::FindStunEndpointJoin(*enetClient, stunEndpoint))
+		if (!StunClient::FindPublicIP(*enetClient, ip, port))
 		{
 			PushGuiMessage(
 				"type", "netstatus",
@@ -256,10 +271,48 @@ bool CNetClient::TryToConnect(const CStr& hostJID)
 			return false;
 		}
 
-		g_XmppClient->SendStunEndpointToHost(stunEndpoint, hostJID);
+		// If the host is on the same network, we risk failing to connect
+		// on routers that don't support NAT hairpinning/NAT loopback.
+		// To work around that, send again a connection data request, but for internal IP this time.
+		if (ip == m_ServerAddress)
+		{
+			g_XmppClient->SendIqGetConnectionData(m_HostJID, m_Password, m_UserName.ToUTF8(), true);
+			// Return true anyways - we're on a success path here.
+			return true;
+		}
+	}
+	else if (g_XmppClient && localNetwork)
+	{
+		// We may need to punch a hole through the local firewall, so fetch our local IP.
+		// NB: we'll ignore failures here, and hope that the firewall will be open to connection
+		// if we fail to fetch the local IP (which is unlikely anyways).
+		if (!StunClient::FindLocalIP(ip))
+			ip = "";
+		// Check if we're hosting on localhost, and if so, explicitly use that
+		// (this circumvents, at least, the 'block all incoming connections' setting
+		// on the MacOS firewall).
+		if (ip == m_ServerAddress)
+		{
+			m_ServerAddress = "127.0.0.1";
+			ip = "";
+		}
+		port = enetClient->address.port;
+	}
 
+	LOGMESSAGE("NetClient: connecting to server at %s:%i", m_ServerAddress, m_ServerPort);
+
+	if (!ip.empty())
+	{
+		// UDP hole-punching
+		// Step 0: send a message, via XMPP, to the server with our external IP & port.
+		g_XmppClient->SendStunEndpointToHost(ip, port, hostJID);
+
+		// Step 1b: Wait some time - we need the host to receive the stun endpoint and start punching a hole themselves before
+		// we try to establish the connection below.
 		SDL_Delay(1000);
 
+		// Step 2: Send a message ourselves to the server so that the NAT, if any, routes incoming trafic correctly.
+		// TODO: verify if this step is necessary, since we'll try and connect anyways below.
 		StunClient::SendHolePunchingMessages(*enetClient, m_ServerAddress, m_ServerPort);
 	}
 
@@ -329,9 +382,9 @@ void CNetClient::CheckServerConnection()
 		return;
 	}
 
-	// Report if we have a bad ping to the server
+	// Report if we have a bad ping to the server.
 	u32 meanRTT = m_Session->GetMeanRTT();
-	if (meanRTT > DEFAULT_TURN_LENGTH_MP)
+	if (meanRTT > NETWORK_BAD_PING)
 	{
 		PushGuiMessage(
 			"type", "netwarn",
@@ -363,7 +416,7 @@ std::string CNetClient::TestReadGuiMessages()
 		GuiPoll(&msg);
 		if (msg.isUndefined())
 			break;
-		r += GetScriptInterface().ToString(&msg) + "\n";
+		r += Script::ToString(rq, &msg) + "\n";
 	}
 	return r;
 }
@@ -378,20 +431,20 @@ void CNetClient::PostPlayerAssignmentsToScript()
 	ScriptRequest rq(GetScriptInterface());
 
 	JS::RootedValue newAssignments(rq.cx);
-	ScriptInterface::CreateObject(rq, &newAssignments);
+	Script::CreateObject(rq, &newAssignments);
 
 	for (const std::pair<const CStr, PlayerAssignment>& p : m_PlayerAssignments)
 	{
 		JS::RootedValue assignment(rq.cx);
 
-		ScriptInterface::CreateObject(
+		Script::CreateObject(
 			rq,
 			&assignment,
 			"name", p.second.m_Name,
 			"player", p.second.m_PlayerID,
 			"status", p.second.m_Status);
 
-		GetScriptInterface().SetProperty(newAssignments, p.first.c_str(), assignment);
+		Script::SetProperty(rq, newAssignments, p.first.c_str(), assignment);
 	}
 
 	PushGuiMessage(
@@ -461,9 +514,10 @@ void CNetClient::SendClearAllReadyMessage()
 	SendMessage(&clearAllReady);
 }
 
-void CNetClient::SendStartGameMessage()
+void CNetClient::SendStartGameMessage(const CStr& initAttribs)
 {
 	CGameStartMessage gameStart;
+	gameStart.m_InitAttributes = initAttribs;
 	SendMessage(&gameStart);
 }
 
@@ -576,7 +630,7 @@ void CNetClient::SendAuthenticateMessage()
 	CAuthenticateMessage authenticate;
 	authenticate.m_Name = m_UserName;
 	authenticate.m_Password = m_Password;
-	authenticate.m_IsLocalClient = m_IsLocalClient;
+	authenticate.m_ControllerSecret = m_ControllerSecret;
 	SendMessage(&authenticate);
 }
 
@@ -619,8 +673,8 @@ bool CNetClient::OnHandshakeResponse(void* context, CFsmEvent* event)
 
 	if (message->m_Flags & PS_NETWORK_FLAG_REQUIRE_LOBBYAUTH)
 	{
-		if (g_XmppClient && !client->m_HostingPlayerName.empty())
-			g_XmppClient->SendIqLobbyAuth(client->m_HostingPlayerName, client->m_GUID);
+		if (g_XmppClient && !client->m_HostJID.empty())
+			g_XmppClient->SendIqLobbyAuth(client->m_HostJID, client->m_GUID);
 		else
 		{
 			client->PushGuiMessage(
@@ -657,6 +711,7 @@ bool CNetClient::OnAuthenticate(void* context, CFsmEvent* event)
 
 	client->m_HostID = message->m_HostID;
 	client->m_Rejoin = message->m_Code == ARC_OK_REJOINING;
+	client->m_IsController = message->m_IsController;
 
 	client->PushGuiMessage(
 		"type", "netstatus",
@@ -703,8 +758,6 @@ bool CNetClient::OnGameSetup(void* context, CFsmEvent* event)
 	CNetClient* client = static_cast<CNetClient*>(context);
 	CGameSetupMessage* message = static_cast<CGameSetupMessage*>(event->GetParamRef());
 
-	client->m_GameAttributes = message->m_Data;
-
 	client->PushGuiMessage(
 		"type", "gamesetup",
 		"data", message->m_Data);
@@ -745,6 +798,7 @@ bool CNetClient::OnGameStart(void* context, CFsmEvent* event)
 	ENSURE(event->GetType() == (uint)NMT_GAME_START);
 
 	CNetClient* client = static_cast<CNetClient*>(context);
+	CGameStartMessage* message = static_cast<CGameStartMessage*>(event->GetParamRef());
 
 	// Find the player assigned to our GUID
 	int player = -1;
@@ -754,10 +808,17 @@ bool CNetClient::OnGameStart(void* context, CFsmEvent* event)
 	client->m_ClientTurnManager = new CNetClientTurnManager(
 			*client->m_Game->GetSimulation2(), *client, client->m_HostID, client->m_Game->GetReplayLogger());
 
-	client->m_Game->SetPlayerID(player);
-	client->m_Game->StartGame(&client->m_GameAttributes, "");
+	// Parse init attributes.
+	const ScriptInterface& scriptInterface = client->m_Game->GetSimulation2()->GetScriptInterface();
+	ScriptRequest rq(scriptInterface);
+	JS::RootedValue initAttribs(rq.cx);
+	Script::ParseJSON(rq, message->m_InitAttributes, &initAttribs);
 
-	client->PushGuiMessage("type", "start");
+	client->m_Game->SetPlayerID(player);
+	client->m_Game->StartGame(&initAttribs, "");
+
+	client->PushGuiMessage("type", "start",
+						   "initAttributes", initAttribs);
 
 	return true;
 }
@@ -768,9 +829,11 @@ bool CNetClient::OnJoinSyncStart(void* context, CFsmEvent* event)
 
 	CNetClient* client = static_cast<CNetClient*>(context);
 
+	CJoinSyncStartMessage* joinSyncStartMessage = (CJoinSyncStartMessage*)event->GetParamRef();
+
 	// The server wants us to start downloading the game state from it, so do so
 	client->m_Session->GetFileTransferer().StartTask(
-		shared_ptr<CNetFileReceiveTask>(new CNetFileReceiveTask_ClientRejoin(*client))
+		std::shared_ptr<CNetFileReceiveTask>(new CNetFileReceiveTask_ClientRejoin(*client, joinSyncStartMessage->m_InitAttributes))
 	);
 
 	return true;
@@ -851,7 +914,7 @@ bool CNetClient::OnClientPerformance(void *context, CFsmEvent* event)
 	// Display warnings for other clients with bad ping
 	for (size_t i = 0; i < message->m_Clients.size(); ++i)
 	{
-		if (message->m_Clients[i].m_MeanRTT < DEFAULT_TURN_LENGTH_MP || message->m_Clients[i].m_GUID == client->m_GUID)
+		if (message->m_Clients[i].m_MeanRTT < NETWORK_BAD_PING || message->m_Clients[i].m_GUID == client->m_GUID)
 			continue;
 
 		client->PushGuiMessage(
