@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2021 Wildfire Games.
  *
  * Permission is hereby granted, free of charge, to any person obtaining
  * a copy of this software and associated documentation files (the
@@ -31,12 +31,15 @@
 #include "maths/MD5.h"
 #include "ps/CLogger.h"
 #include "ps/ConfigDB.h"
+#include "ps/GameSetup/CmdLineArgs.h"
 #include "ps/GameSetup/Paths.h"
 #include "ps/Mod.h"
 #include "ps/ModInstaller.h"
 #include "ps/Util.h"
 #include "scriptinterface/ScriptConversions.h"
-#include "scriptinterface/ScriptInterface.h"
+#include "scriptinterface/ScriptContext.h"
+#include "scriptinterface/ScriptRequest.h"
+#include "scriptinterface/JSON.h"
 
 #include <boost/algorithm/string.hpp>
 #include <iomanip>
@@ -296,7 +299,7 @@ void ModIo::StartListMods()
 	m_DownloadProgressData.status = DownloadProgressStatus::LISTING;
 }
 
-void ModIo::StartDownloadMod(size_t idx)
+void ModIo::StartDownloadMod(u32 idx)
 {
 	// Don't start such a request during active downloads.
 	if (m_DownloadProgressData.status == DownloadProgressStatus::GAMEID ||
@@ -307,7 +310,7 @@ void ModIo::StartDownloadMod(size_t idx)
 	if (idx >= m_ModData.size())
 		return;
 
-	const Paths paths(g_args);
+	const Paths paths(g_CmdLineArgs);
 	const OsPath modUserPath = paths.UserData()/"mods";
 	const OsPath modPath = modUserPath/m_ModData[idx].properties["name_id"];
 	if (!DirectoryExists(modPath) && INFO::OK != CreateDirectories(modPath, 0700, false))
@@ -480,9 +483,12 @@ bool ModIo::AdvanceRequest(const ScriptInterface& scriptInterface)
 		m_DownloadProgressData.status = DownloadProgressStatus::SUCCESS;
 
 		{
-			Paths paths(g_args);
+			Paths paths(g_CmdLineArgs);
 			CModInstaller installer(paths.UserData() / "mods", paths.Cache());
-			installer.Install(m_DownloadFilePath, g_ScriptContext, false);
+			CModInstaller::ModInstallationResult result = installer.Install(m_DownloadFilePath, g_ScriptContext, false);
+			if (result != CModInstaller::ModInstallationResult::SUCCESS)
+				LOGERROR("Failed to install '%s'", m_DownloadFilePath.string8().c_str());
+			g_Mods.UpdateAvailableMods(scriptInterface);
 		}
 		break;
 	default:
@@ -589,7 +595,7 @@ bool ModIo::ParseGameIdResponse(const ScriptInterface& scriptInterface, const st
 
 	JS::RootedValue gameResponse(rq.cx);
 
-	if (!scriptInterface.ParseJSON(responseData, &gameResponse))
+	if (!Script::ParseJSON(rq, responseData, &gameResponse))
 		FAIL("Failed to parse response as JSON.");
 
 	if (!gameResponse.isObject())
@@ -632,7 +638,7 @@ bool ModIo::ParseGameIdResponse(const ScriptInterface& scriptInterface, const st
 		FAIL("id property not a number.");
 
 	id = -1;
-	if (!ScriptInterface::FromJSVal(rq, idProperty, id) || id <= 0)
+	if (!Script::FromJSVal(rq, idProperty, id) || id <= 0)
 		FAIL("Invalid id.");
 
 	return true;
@@ -660,7 +666,7 @@ bool ModIo::ParseModsResponse(const ScriptInterface& scriptInterface, const std:
 
 	JS::RootedValue modResponse(rq.cx);
 
-	if (!scriptInterface.ParseJSON(responseData, &modResponse))
+	if (!Script::ParseJSON(rq, responseData, &modResponse))
 		FAIL("Failed to parse response as JSON.");
 
 	if (!modResponse.isObject())
@@ -705,7 +711,7 @@ bool ModIo::ParseModsResponse(const ScriptInterface& scriptInterface, const std:
 	for (const std::string& prop : { __VA_ARGS__ }) \
 	{ \
 		std::string val; \
-		if (!ScriptInterface::FromJSProperty(rq, obj, prop.c_str(), val, true)) \
+		if (!Script::FromJSProperty(rq, obj, prop.c_str(), val, true)) \
 		{ \
 			ok  = false; \
 			copyStringError = "Failed to get " + prop + " from " + #obj + "."; \
@@ -745,21 +751,21 @@ bool ModIo::ParseModsResponse(const ScriptInterface& scriptInterface, const std:
 
 		// Parse metadata_blob (sig+deps)
 		std::string metadata_blob;
-		if (!ScriptInterface::FromJSProperty(rq, modFile, "metadata_blob", metadata_blob, true))
+		if (!Script::FromJSProperty(rq, modFile, "metadata_blob", metadata_blob, true))
 			INVALIDATE_DATA_AND_CONTINUE("Failed to get metadata_blob from modFile.");
 
 		JS::RootedValue metadata(rq.cx);
-		if (!scriptInterface.ParseJSON(metadata_blob, &metadata))
+		if (!Script::ParseJSON(rq, metadata_blob, &metadata))
 			INVALIDATE_DATA_AND_CONTINUE("Failed to parse metadata_blob as JSON.");
 
 		if (!metadata.isObject())
 			INVALIDATE_DATA_AND_CONTINUE("metadata_blob is not decoded as an object.");
 
-		if (!ScriptInterface::FromJSProperty(rq, metadata, "dependencies", data.dependencies, true))
+		if (!Script::FromJSProperty(rq, metadata, "dependencies", data.dependencies, true))
 			INVALIDATE_DATA_AND_CONTINUE("Failed to get dependencies from metadata_blob.");
 
 		std::vector<std::string> minisigs;
-		if (!ScriptInterface::FromJSProperty(rq, metadata, "minisigs", minisigs, true))
+		if (!Script::FromJSProperty(rq, metadata, "minisigs", minisigs, true))
 			INVALIDATE_DATA_AND_CONTINUE("Failed to get minisigs from metadata_blob.");
 
 		// Check we did find a valid matching signature.
@@ -847,7 +853,7 @@ bool ModIo::ParseSignature(const std::vector<std::string>& minisigs, SigStruct&
 		return true;
 	}
 
-	return false;
+	FAIL("Invalid signature.");
 #undef CLEANUP
 }
 
