@@ -1,4 +1,4 @@
-/* Copyright (C) 2021 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -27,17 +27,22 @@
 #include "graphics/Terrain.h"
 #include "graphics/TerrainTextureEntry.h"
 #include "graphics/TextRenderer.h"
+#include "graphics/TextureManager.h"
 #include "lib/allocators/DynamicArena.h"
 #include "lib/allocators/STLAllocators.h"
 #include "maths/MathUtil.h"
 #include "ps/CLogger.h"
+#include "ps/CStrInternStatic.h"
 #include "ps/Game.h"
 #include "ps/GameSetup/Config.h"
 #include "ps/Profile.h"
 #include "ps/Pyrogenesis.h"
+#include "ps/VideoMode.h"
 #include "ps/World.h"
 #include "renderer/AlphaMapCalculator.h"
+#include "renderer/DebugRenderer.h"
 #include "renderer/Renderer.h"
+#include "renderer/SceneRenderer.h"
 #include "renderer/TerrainRenderer.h"
 #include "renderer/WaterManager.h"
 #include "simulation2/components/ICmpWaterManager.h"
@@ -60,12 +65,7 @@ const ssize_t BlendOffsets[9][2] = {
 };
 
 CPatchRData::CPatchRData(CPatch* patch, CSimulation2* simulation) :
-	m_Patch(patch), m_VBSides(),
-	m_VBBase(), m_VBBaseIndices(),
-	m_VBBlends(), m_VBBlendIndices(),
-	m_VBWater(), m_VBWaterIndices(),
-	m_VBWaterShore(), m_VBWaterIndicesShore(),
-	m_Simulation(simulation)
+	m_Patch(patch), m_Simulation(simulation)
 {
 	ENSURE(patch);
 	Build();
@@ -281,14 +281,20 @@ void CPatchRData::BuildBlends()
 	{
 		// Construct vertex buffer
 
-		m_VBBlends = g_VBMan.AllocateChunk(sizeof(SBlendVertex), blendVertices.size(), GL_STATIC_DRAW, GL_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::TERRAIN);
+		m_VBBlends = g_VBMan.AllocateChunk(
+			sizeof(SBlendVertex), blendVertices.size(),
+			Renderer::Backend::IBuffer::Type::VERTEX, false,
+			nullptr, CVertexBufferManager::Group::TERRAIN);
 		m_VBBlends->m_Owner->UpdateChunkVertices(m_VBBlends.Get(), &blendVertices[0]);
 
 		// Update the indices to include the base offset of the vertex data
 		for (size_t k = 0; k < blendIndices.size(); ++k)
 			blendIndices[k] += static_cast<u16>(m_VBBlends->m_Index);
 
-		m_VBBlendIndices = g_VBMan.AllocateChunk(sizeof(u16), blendIndices.size(), GL_STATIC_DRAW, GL_ELEMENT_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::TERRAIN);
+		m_VBBlendIndices = g_VBMan.AllocateChunk(
+			sizeof(u16), blendIndices.size(),
+			Renderer::Backend::IBuffer::Type::INDEX, false,
+			nullptr, CVertexBufferManager::Group::TERRAIN);
 		m_VBBlendIndices->m_Owner->UpdateChunkVertices(m_VBBlendIndices.Get(), &blendIndices[0]);
 	}
 }
@@ -489,7 +495,9 @@ void CPatchRData::BuildIndices()
 	ENSURE(indices.size());
 
 	// Construct vertex buffer
-	m_VBBaseIndices = g_VBMan.AllocateChunk(sizeof(u16), indices.size(), GL_STATIC_DRAW, GL_ELEMENT_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::TERRAIN);
+	m_VBBaseIndices = g_VBMan.AllocateChunk(
+		sizeof(u16), indices.size(),
+		Renderer::Backend::IBuffer::Type::INDEX, false, nullptr, CVertexBufferManager::Group::TERRAIN);
 	m_VBBaseIndices->m_Owner->UpdateChunkVertices(m_VBBaseIndices.Get(), &indices[0]);
 }
 
@@ -532,7 +540,12 @@ void CPatchRData::BuildVertices()
 
 	// upload to vertex buffer
 	if (!m_VBBase)
-		m_VBBase = g_VBMan.AllocateChunk(sizeof(SBaseVertex), vsize * vsize, GL_STATIC_DRAW, GL_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::TERRAIN);
+	{
+		m_VBBase = g_VBMan.AllocateChunk(
+			sizeof(SBaseVertex), vsize * vsize,
+			Renderer::Backend::IBuffer::Type::VERTEX, false,
+			nullptr, CVertexBufferManager::Group::TERRAIN);
+	}
 
 	m_VBBase->m_Owner->UpdateChunkVertices(m_VBBase.Get(), &vertices[0]);
 }
@@ -569,17 +582,24 @@ void CPatchRData::BuildSide(std::vector<SSideVertex>& vertices, CPatchSideFlags
 		v1.m_Position = pos;
 		v1.m_Position.Y = 0;
 
-		// If this is the start of this tristrip, but we've already got a partial
-		// tristrip, add a couple of degenerate triangles to join the strips properly
-		if (k == 0 && !vertices.empty())
+		if (k == 0)
 		{
-			vertices.push_back(vertices.back());
-			vertices.push_back(v1);
+			vertices.emplace_back(v1);
+			vertices.emplace_back(v0);
+		}
+		if (k > 0)
+		{
+			const size_t lastIndex = vertices.size() - 1;
+			vertices.emplace_back(v1);
+			vertices.emplace_back(vertices[lastIndex]);
+			vertices.emplace_back(v0);
+			vertices.emplace_back(v1);
+			if (k + 1 < vsize)
+			{
+				vertices.emplace_back(v1);
+				vertices.emplace_back(v0);
+			}
 		}
-
-		// Now add the new triangles
-		vertices.push_back(v1);
-		vertices.push_back(v0);
 	}
 }
 
@@ -614,7 +634,12 @@ void CPatchRData::BuildSides()
 		return;
 
 	if (!m_VBSides)
-		m_VBSides = g_VBMan.AllocateChunk(sizeof(SSideVertex), sideVertices.size(), GL_STATIC_DRAW, GL_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::DEFAULT);
+	{
+		m_VBSides = g_VBMan.AllocateChunk(
+			sizeof(SSideVertex), sideVertices.size(),
+			Renderer::Backend::IBuffer::Type::VERTEX, false,
+			nullptr, CVertexBufferManager::Group::DEFAULT);
+	}
 	m_VBSides->m_Owner->UpdateChunkVertices(m_VBSides.Get(), &sideVertices[0]);
 }
 
@@ -644,8 +669,6 @@ void CPatchRData::Update(CSimulation2* simulation)
 	}
 }
 
-// Types used for glMultiDrawElements batching:
-
 // To minimise the cost of memory allocations, everything used for computing
 // batches uses a arena allocator. (All allocations are short-lived so we can
 // just throw away the whole arena at the end of each frame.)
@@ -677,7 +700,7 @@ typename M::mapped_type& PooledPairGet(M& m, const typename M::key_type& k, Aren
 }
 
 // Each multidraw batch has a list of index counts, and a list of pointers-to-first-indexes
-using BatchElements = std::pair<std::vector<GLint, ProxyAllocator<GLint, Arena>>, std::vector<void*, ProxyAllocator<void*, Arena>>>;
+using BatchElements = std::pair<std::vector<u32, ProxyAllocator<u32, Arena>>, std::vector<u32, ProxyAllocator<u32, Arena>>>;
 
 // Group batches by index buffer
 using IndexBufferBatches = PooledBatchMap<CVertexBuffer*, BatchElements>;
@@ -689,12 +712,14 @@ using VertexBufferBatches = PooledBatchMap<CVertexBuffer*, IndexBufferBatches>;
 using TextureBatches = PooledBatchMap<CTerrainTextureEntry*, VertexBufferBatches>;
 
 // Group batches by shaders.
-using ShaderTechniqueBatches = PooledBatchMap<CShaderTechniquePtr, TextureBatches>;
+using ShaderTechniqueBatches = PooledBatchMap<std::pair<CStrIntern, CShaderDefines>, TextureBatches>;
 
 void CPatchRData::RenderBases(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext,
 	const std::vector<CPatchRData*>& patches, const CShaderDefines& context, ShadowMap* shadow)
 {
 	PROFILE3("render terrain bases");
+	GPU_SCOPED_LABEL(deviceCommandContext, "Render terrain bases");
 
 	Arena arena;
 
@@ -715,13 +740,11 @@ void CPatchRData::RenderBases(
 				LOGERROR("Terrain renderer failed to load shader effect.\n");
 				continue;
 			}
-			CShaderTechniquePtr techBase = g_Renderer.GetShaderManager().LoadEffect(
-				material.GetShaderEffect(), context, material.GetShaderDefines(0));
 
 			BatchElements& batch = PooledPairGet(
 				PooledMapGet(
 					PooledMapGet(
-						PooledMapGet(batches, techBase, arena),
+						PooledMapGet(batches, std::make_pair(material.GetShaderEffect(), material.GetShaderDefines()), arena),
 						splat.m_Texture, arena
 					),
 					patch->m_VBBase->m_Owner, arena
@@ -731,8 +754,7 @@ void CPatchRData::RenderBases(
 
 			batch.first.push_back(splat.m_IndexCount);
 
-			u8* indexBase = patch->m_VBBaseIndices->m_Owner->GetBindAddress();
-			batch.second.push_back(indexBase + sizeof(u16)*(patch->m_VBBaseIndices->m_Index + splat.m_IndexStart));
+			batch.second.push_back(patch->m_VBBaseIndices->m_Index + splat.m_IndexStart);
 		}
 	}
 
@@ -741,69 +763,98 @@ void CPatchRData::RenderBases(
 	// Render each batch
 	for (ShaderTechniqueBatches::iterator itTech = batches.begin(); itTech != batches.end(); ++itTech)
 	{
-		const CShaderTechniquePtr& techBase = itTech->first;
+		CShaderDefines defines = context;
+		defines.SetMany(itTech->first.second);
+		CShaderTechniquePtr techBase = g_Renderer.GetShaderManager().LoadEffect(
+			itTech->first.first, defines);
+
 		const int numPasses = techBase->GetNumPasses();
 		for (int pass = 0; pass < numPasses; ++pass)
 		{
-			techBase->BeginPass(pass);
-			const CShaderProgramPtr& shader = techBase->GetShader(pass);
-			TerrainRenderer::PrepareShader(shader, shadow);
+			deviceCommandContext->SetGraphicsPipelineState(
+				techBase->GetGraphicsPipelineStateDesc(pass));
+			deviceCommandContext->BeginPass();
+			Renderer::Backend::IShaderProgram* shader = techBase->GetShader(pass);
+			TerrainRenderer::PrepareShader(deviceCommandContext, shader, shadow);
+
+			const int32_t baseTexBindingSlot =
+				shader->GetBindingSlot(str_baseTex);
+			const int32_t textureTransformBindingSlot =
+				shader->GetBindingSlot(str_textureTransform);
 
 			TextureBatches& textureBatches = itTech->second;
 			for (TextureBatches::iterator itt = textureBatches.begin(); itt != textureBatches.end(); ++itt)
 			{
-				if (itt->first->GetMaterial().GetSamplers().size() != 0)
+				if (!itt->first->GetMaterial().GetSamplers().empty())
 				{
-					const CMaterial::SamplersVector& samplers = itt->first->GetMaterial().GetSamplers();
+					const CMaterial::SamplersVector& samplers =
+						itt->first->GetMaterial().GetSamplers();
 					for(const CMaterial::TextureSampler& samp : samplers)
-						shader->BindTexture(samp.Name, samp.Sampler);
+						samp.Sampler->UploadBackendTextureIfNeeded(deviceCommandContext);
+					for(const CMaterial::TextureSampler& samp : samplers)
+					{
+						deviceCommandContext->SetTexture(
+							shader->GetBindingSlot(samp.Name),
+							samp.Sampler->GetBackendTexture());
+					}
 
-					itt->first->GetMaterial().GetStaticUniforms().BindUniforms(shader);
+					itt->first->GetMaterial().GetStaticUniforms().BindUniforms(
+						deviceCommandContext, shader);
 
 					float c = itt->first->GetTextureMatrix()[0];
 					float ms = itt->first->GetTextureMatrix()[8];
-					shader->Uniform(str_textureTransform, c, ms, -ms, 0.f);
+					deviceCommandContext->SetUniform(
+						textureTransformBindingSlot, c, ms);
 				}
 				else
 				{
-					shader->BindTexture(str_baseTex, g_Renderer.GetTextureManager().GetErrorTexture());
+					deviceCommandContext->SetTexture(
+						baseTexBindingSlot,
+						g_Renderer.GetTextureManager().GetErrorTexture()->GetBackendTexture());
 				}
 
 				for (VertexBufferBatches::iterator itv = itt->second.begin(); itv != itt->second.end(); ++itv)
 				{
-					GLsizei stride = sizeof(SBaseVertex);
-					SBaseVertex *base = (SBaseVertex *)itv->first->Bind();
-					shader->VertexPointer(3, GL_FLOAT, stride, &base->m_Position[0]);
-					shader->NormalPointer(GL_FLOAT, stride, &base->m_Normal[0]);
-					shader->TexCoordPointer(GL_TEXTURE0, 3, GL_FLOAT, stride, &base->m_Position[0]);
+					itv->first->UploadIfNeeded(deviceCommandContext);
 
-					shader->AssertPointersBound();
+					const uint32_t stride = sizeof(SBaseVertex);
+
+					deviceCommandContext->SetVertexAttributeFormat(
+						Renderer::Backend::VertexAttributeStream::POSITION,
+						Renderer::Backend::Format::R32G32B32_SFLOAT,
+						offsetof(SBaseVertex, m_Position), stride,
+						Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+					deviceCommandContext->SetVertexAttributeFormat(
+						Renderer::Backend::VertexAttributeStream::NORMAL,
+						Renderer::Backend::Format::R32G32B32_SFLOAT,
+						offsetof(SBaseVertex, m_Normal), stride,
+						Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+					deviceCommandContext->SetVertexAttributeFormat(
+						Renderer::Backend::VertexAttributeStream::UV0,
+						Renderer::Backend::Format::R32G32B32_SFLOAT,
+						offsetof(SBaseVertex, m_Position), stride,
+						Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+
+					deviceCommandContext->SetVertexBuffer(0, itv->first->GetBuffer());
 
 					for (IndexBufferBatches::iterator it = itv->second.begin(); it != itv->second.end(); ++it)
 					{
-						it->first->Bind();
+						it->first->UploadIfNeeded(deviceCommandContext);
+						deviceCommandContext->SetIndexBuffer(it->first->GetBuffer());
 
 						BatchElements& batch = it->second;
 
-						if (!g_Renderer.m_SkipSubmit)
-						{
-							// Don't use glMultiDrawElements here since it doesn't have a significant
-							// performance impact and it suffers from various driver bugs (e.g. it breaks
-							// in Mesa 7.10 swrast with index VBOs)
-							for (size_t i = 0; i < batch.first.size(); ++i)
-								glDrawElements(GL_TRIANGLES, batch.first[i], GL_UNSIGNED_SHORT, batch.second[i]);
-						}
+						for (size_t i = 0; i < batch.first.size(); ++i)
+							deviceCommandContext->DrawIndexed(batch.second[i], batch.first[i], 0);
 
 						g_Renderer.m_Stats.m_DrawCalls++;
 						g_Renderer.m_Stats.m_TerrainTris += std::accumulate(batch.first.begin(), batch.first.end(), 0) / 3;
 					}
 				}
 			}
-			techBase->EndPass();
+			deviceCommandContext->EndPass();
 		}
 	}
-
-	CVertexBuffer::Unbind();
 }
 
 /**
@@ -839,9 +890,11 @@ struct SBlendStackItem
 };
 
 void CPatchRData::RenderBlends(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext,
 	const std::vector<CPatchRData*>& patches, const CShaderDefines& context, ShadowMap* shadow)
 {
 	PROFILE3("render terrain blends");
+	GPU_SCOPED_LABEL(deviceCommandContext, "Render terrain blends");
 
 	Arena arena;
 
@@ -896,8 +949,7 @@ void CPatchRData::RenderBlends(
 					BatchElements& batch = PooledPairGet(PooledMapGet(batches.back().m_Batches, vertices->m_Owner, arena), indices->m_Owner, arena);
 					batch.first.push_back(splats.back().m_IndexCount);
 
-		 			u8* indexBase = indices->m_Owner->GetBindAddress();
-		 			batch.second.push_back(indexBase + sizeof(u16)*(indices->m_Index + splats.back().m_IndexStart));
+		 			batch.second.push_back(indices->m_Index + splats.back().m_IndexStart);
 
 					splats.pop_back();
 				}
@@ -924,19 +976,18 @@ void CPatchRData::RenderBlends(
 		layer.m_Texture = bestTex;
 		if (!bestTex->GetMaterial().GetSamplers().empty())
 		{
+			CShaderDefines defines = contextBlend;
+			defines.SetMany(bestTex->GetMaterial().GetShaderDefines());
 			layer.m_ShaderTech = g_Renderer.GetShaderManager().LoadEffect(
-				bestTex->GetMaterial().GetShaderEffect(), contextBlend, bestTex->GetMaterial().GetShaderDefines(0));
+				bestTex->GetMaterial().GetShaderEffect(), defines);
 		}
 		batches.push_back(layer);
 	}
 
 	PROFILE_END("compute batches");
 
-	glEnable(GL_BLEND);
-	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-
 	CVertexBuffer* lastVB = nullptr;
-	CShaderProgramPtr previousShader;
+	Renderer::Backend::IShaderProgram* previousShader = nullptr;
 	for (BatchesStack::iterator itTechBegin = batches.begin(), itTechEnd = batches.begin(); itTechBegin != batches.end(); itTechBegin = itTechEnd)
 	{
 		while (itTechEnd != batches.end() && itTechEnd->m_ShaderTech == itTechBegin->m_ShaderTech)
@@ -946,11 +997,29 @@ void CPatchRData::RenderBlends(
 		const int numPasses = techBase->GetNumPasses();
 		for (int pass = 0; pass < numPasses; ++pass)
 		{
-			techBase->BeginPass(pass);
-			const CShaderProgramPtr& shader = techBase->GetShader(pass);
-			TerrainRenderer::PrepareShader(shader, shadow);
+			Renderer::Backend::GraphicsPipelineStateDesc pipelineStateDesc =
+				techBase->GetGraphicsPipelineStateDesc(pass);
+			pipelineStateDesc.blendState.enabled = true;
+			pipelineStateDesc.blendState.srcColorBlendFactor = pipelineStateDesc.blendState.srcAlphaBlendFactor =
+				Renderer::Backend::BlendFactor::SRC_ALPHA;
+			pipelineStateDesc.blendState.dstColorBlendFactor = pipelineStateDesc.blendState.dstAlphaBlendFactor =
+				Renderer::Backend::BlendFactor::ONE_MINUS_SRC_ALPHA;
+			pipelineStateDesc.blendState.colorBlendOp = pipelineStateDesc.blendState.alphaBlendOp =
+				Renderer::Backend::BlendOp::ADD;
+			deviceCommandContext->SetGraphicsPipelineState(pipelineStateDesc);
+			deviceCommandContext->BeginPass();
 
-			Handle lastBlendTex = 0;
+			Renderer::Backend::IShaderProgram* shader = techBase->GetShader(pass);
+			TerrainRenderer::PrepareShader(deviceCommandContext, shader, shadow);
+
+			Renderer::Backend::ITexture* lastBlendTex = nullptr;
+
+			const int32_t baseTexBindingSlot =
+				shader->GetBindingSlot(str_baseTex);
+			const int32_t blendTexBindingSlot =
+				shader->GetBindingSlot(str_blendTex);
+			const int32_t textureTransformBindingSlot =
+				shader->GetBindingSlot(str_textureTransform);
 
 			for (BatchesStack::iterator itt = itTechBegin; itt != itTechEnd; ++itt)
 			{
@@ -961,24 +1030,33 @@ void CPatchRData::RenderBlends(
 				{
 					const CMaterial::SamplersVector& samplers = itt->m_Texture->GetMaterial().GetSamplers();
 					for (const CMaterial::TextureSampler& samp : samplers)
-						shader->BindTexture(samp.Name, samp.Sampler);
+						samp.Sampler->UploadBackendTextureIfNeeded(deviceCommandContext);
+					for (const CMaterial::TextureSampler& samp : samplers)
+					{
+						deviceCommandContext->SetTexture(
+							shader->GetBindingSlot(samp.Name),
+							samp.Sampler->GetBackendTexture());
+					}
 
-					Handle currentBlendTex = itt->m_Texture->m_TerrainAlpha->second.m_hCompositeAlphaMap;
+					Renderer::Backend::ITexture* currentBlendTex = itt->m_Texture->m_TerrainAlpha->second.m_CompositeAlphaMap.get();
 					if (currentBlendTex != lastBlendTex)
 					{
-						shader->BindTexture(str_blendTex, currentBlendTex);
+						deviceCommandContext->SetTexture(
+							blendTexBindingSlot, currentBlendTex);
 						lastBlendTex = currentBlendTex;
 					}
 
-					itt->m_Texture->GetMaterial().GetStaticUniforms().BindUniforms(shader);
+					itt->m_Texture->GetMaterial().GetStaticUniforms().BindUniforms(deviceCommandContext, shader);
 
 					float c = itt->m_Texture->GetTextureMatrix()[0];
 					float ms = itt->m_Texture->GetTextureMatrix()[8];
-					shader->Uniform(str_textureTransform, c, ms, -ms, 0.f);
+					deviceCommandContext->SetUniform(
+						textureTransformBindingSlot, c, ms);
 				}
 				else
 				{
-					shader->BindTexture(str_baseTex, g_Renderer.GetTextureManager().GetErrorTexture());
+					deviceCommandContext->SetTexture(
+						baseTexBindingSlot, g_Renderer.GetTextureManager().GetErrorTexture()->GetBackendTexture());
 				}
 
 				for (VertexBufferBatches::iterator itv = itt->m_Batches.begin(); itv != itt->m_Batches.end(); ++itv)
@@ -988,28 +1066,44 @@ void CPatchRData::RenderBlends(
 					{
 						lastVB = itv->first;
 						previousShader = shader;
-						GLsizei stride = sizeof(SBlendVertex);
-						SBlendVertex *base = (SBlendVertex *)itv->first->Bind();
 
-						shader->VertexPointer(3, GL_FLOAT, stride, &base->m_Position[0]);
-						shader->NormalPointer(GL_FLOAT, stride, &base->m_Normal[0]);
-						shader->TexCoordPointer(GL_TEXTURE0, 3, GL_FLOAT, stride, &base->m_Position[0]);
-						shader->TexCoordPointer(GL_TEXTURE1, 2, GL_FLOAT, stride, &base->m_AlphaUVs[0]);
+						itv->first->UploadIfNeeded(deviceCommandContext);
+
+						const uint32_t stride = sizeof(SBlendVertex);
+
+						deviceCommandContext->SetVertexAttributeFormat(
+							Renderer::Backend::VertexAttributeStream::POSITION,
+							Renderer::Backend::Format::R32G32B32_SFLOAT,
+							offsetof(SBlendVertex, m_Position), stride,
+							Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+						deviceCommandContext->SetVertexAttributeFormat(
+							Renderer::Backend::VertexAttributeStream::NORMAL,
+							Renderer::Backend::Format::R32G32B32_SFLOAT,
+							offsetof(SBlendVertex, m_Normal), stride,
+							Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+						deviceCommandContext->SetVertexAttributeFormat(
+							Renderer::Backend::VertexAttributeStream::UV0,
+							Renderer::Backend::Format::R32G32B32_SFLOAT,
+							offsetof(SBlendVertex, m_Position), stride,
+							Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+						deviceCommandContext->SetVertexAttributeFormat(
+							Renderer::Backend::VertexAttributeStream::UV1,
+							Renderer::Backend::Format::R32G32_SFLOAT,
+							offsetof(SBlendVertex, m_AlphaUVs), stride,
+							Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+
+						deviceCommandContext->SetVertexBuffer(0, itv->first->GetBuffer());
 					}
 
-					shader->AssertPointersBound();
-
 					for (IndexBufferBatches::iterator it = itv->second.begin(); it != itv->second.end(); ++it)
 					{
-						it->first->Bind();
+						it->first->UploadIfNeeded(deviceCommandContext);
+						deviceCommandContext->SetIndexBuffer(it->first->GetBuffer());
 
 						BatchElements& batch = it->second;
 
-						if (!g_Renderer.m_SkipSubmit)
-						{
-							for (size_t i = 0; i < batch.first.size(); ++i)
-								glDrawElements(GL_TRIANGLES, batch.first[i], GL_UNSIGNED_SHORT, batch.second[i]);
-						}
+						for (size_t i = 0; i < batch.first.size(); ++i)
+							deviceCommandContext->DrawIndexed(batch.second[i], batch.first[i], 0);
 
 						g_Renderer.m_Stats.m_DrawCalls++;
 						g_Renderer.m_Stats.m_BlendSplats++;
@@ -1017,27 +1111,25 @@ void CPatchRData::RenderBlends(
 					}
 				}
 			}
-			techBase->EndPass();
+			deviceCommandContext->EndPass();
 		}
 	}
-
-	glDisable(GL_BLEND);
-
-	CVertexBuffer::Unbind();
 }
 
-void CPatchRData::RenderStreams(const std::vector<CPatchRData*>& patches, const CShaderProgramPtr& shader, int streamflags)
+void CPatchRData::RenderStreams(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext,
+	const std::vector<CPatchRData*>& patches, const bool bindPositionAsTexCoord)
 {
 	PROFILE3("render terrain streams");
 
 	// Each batch has a list of index counts, and a list of pointers-to-first-indexes
-	using StreamBatchElements = std::pair<std::vector<GLint>, std::vector<void*> > ;
+	using StreamBatchElements = std::pair<std::vector<u32>, std::vector<u32>>;
 
 	// Group batches by index buffer
-	using StreamIndexBufferBatches = std::map<CVertexBuffer*, StreamBatchElements> ;
+	using StreamIndexBufferBatches = std::map<CVertexBuffer*, StreamBatchElements>;
 
 	// Group batches by vertex buffer
-	using StreamVertexBufferBatches = std::map<CVertexBuffer*, StreamIndexBufferBatches> ;
+	using StreamVertexBufferBatches = std::map<CVertexBuffer*, StreamIndexBufferBatches>;
 
 	StreamVertexBufferBatches batches;
 
@@ -1050,46 +1142,48 @@ void CPatchRData::RenderStreams(const std::vector<CPatchRData*>& patches, const
 
 		batch.first.push_back(patch->m_VBBaseIndices->m_Count);
 
-		u8* indexBase = patch->m_VBBaseIndices->m_Owner->GetBindAddress();
- 		batch.second.push_back(indexBase + sizeof(u16)*(patch->m_VBBaseIndices->m_Index));
+ 		batch.second.push_back(patch->m_VBBaseIndices->m_Index);
  	}
 
  	PROFILE_END("compute batches");
 
-	ENSURE(!(streamflags & ~(STREAM_POS|STREAM_POSTOUV0|STREAM_POSTOUV1)));
+	const uint32_t stride = sizeof(SBaseVertex);
+
+	deviceCommandContext->SetVertexAttributeFormat(
+		Renderer::Backend::VertexAttributeStream::POSITION,
+		Renderer::Backend::Format::R32G32B32_SFLOAT,
+		offsetof(SBaseVertex, m_Position), stride,
+		Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+	if (bindPositionAsTexCoord)
+	{
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::UV0,
+			Renderer::Backend::Format::R32G32B32_SFLOAT,
+			offsetof(SBaseVertex, m_Position), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+	}
 
  	// Render each batch
 	for (const std::pair<CVertexBuffer* const, StreamIndexBufferBatches>& streamBatch : batches)
 	{
-		GLsizei stride = sizeof(SBaseVertex);
-		SBaseVertex *base = (SBaseVertex *)streamBatch.first->Bind();
+		streamBatch.first->UploadIfNeeded(deviceCommandContext);
 
-		shader->VertexPointer(3, GL_FLOAT, stride, &base->m_Position);
-		if (streamflags & STREAM_POSTOUV0)
-			shader->TexCoordPointer(GL_TEXTURE0, 3, GL_FLOAT, stride, &base->m_Position);
-		if (streamflags & STREAM_POSTOUV1)
-			shader->TexCoordPointer(GL_TEXTURE1, 3, GL_FLOAT, stride, &base->m_Position);
-
-		shader->AssertPointersBound();
+		deviceCommandContext->SetVertexBuffer(0, streamBatch.first->GetBuffer());
 
 		for (const std::pair<CVertexBuffer* const, StreamBatchElements>& batchIndexBuffer : streamBatch.second)
 		{
-			batchIndexBuffer.first->Bind();
+			batchIndexBuffer.first->UploadIfNeeded(deviceCommandContext);
+			deviceCommandContext->SetIndexBuffer(batchIndexBuffer.first->GetBuffer());
 
 			const StreamBatchElements& batch = batchIndexBuffer.second;
 
-			if (!g_Renderer.m_SkipSubmit)
-			{
-				for (size_t i = 0; i < batch.first.size(); ++i)
-					glDrawElements(GL_TRIANGLES, batch.first[i], GL_UNSIGNED_SHORT, batch.second[i]);
-			}
+			for (size_t i = 0; i < batch.first.size(); ++i)
+				deviceCommandContext->DrawIndexed(batch.second[i], batch.first[i], 0);
 
 			g_Renderer.m_Stats.m_DrawCalls++;
 			g_Renderer.m_Stats.m_TerrainTris += std::accumulate(batch.first.begin(), batch.first.end(), 0) / 3;
 		}
 	}
-
-	CVertexBuffer::Unbind();
 }
 
 void CPatchRData::RenderOutline()
@@ -1121,19 +1215,25 @@ void CPatchRData::RenderOutline()
 		line.push_back(pos);
 	}
 
-#if CONFIG2_GLES
-#warning TODO: implement CPatchRData::RenderOutlines for GLES
-#else
-	glVertexPointer(3, GL_FLOAT, sizeof(CVector3D), &line[0]);
-	glDrawArrays(GL_LINE_STRIP, 0, line.size());
-#endif
+	g_Renderer.GetDebugRenderer().DrawLine(line, CColor(0.0f, 0.0f, 1.0f, 1.0f), 0.1f);
 }
 
-void CPatchRData::RenderSides(const std::vector<CPatchRData*>& patches, const CShaderProgramPtr& shader)
+void CPatchRData::RenderSides(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext,
+	const std::vector<CPatchRData*>& patches)
 {
 	PROFILE3("render terrain sides");
+	GPU_SCOPED_LABEL(deviceCommandContext, "Render terrain sides");
 
-	glDisable(GL_CULL_FACE);
+	if (patches.empty())
+		return;
+
+	const uint32_t stride = sizeof(SSideVertex);
+	deviceCommandContext->SetVertexAttributeFormat(
+		Renderer::Backend::VertexAttributeStream::POSITION,
+		Renderer::Backend::Format::R32G32B32_SFLOAT,
+		offsetof(SSideVertex, m_Position), stride,
+		Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
 
 	CVertexBuffer* lastVB = nullptr;
 	for (CPatchRData* patch : patches)
@@ -1144,26 +1244,17 @@ void CPatchRData::RenderSides(const std::vector<CPatchRData*>& patches, const CS
 		if (lastVB != patch->m_VBSides->m_Owner)
 		{
 			lastVB = patch->m_VBSides->m_Owner;
-			SSideVertex *base = (SSideVertex*)patch->m_VBSides->m_Owner->Bind();
+			patch->m_VBSides->m_Owner->UploadIfNeeded(deviceCommandContext);
 
-			// setup data pointers
-			GLsizei stride = sizeof(SSideVertex);
-			shader->VertexPointer(3, GL_FLOAT, stride, &base->m_Position);
+			deviceCommandContext->SetVertexBuffer(0, patch->m_VBSides->m_Owner->GetBuffer());
 		}
 
-		shader->AssertPointersBound();
-
-		if (!g_Renderer.m_SkipSubmit)
-			glDrawArrays(GL_TRIANGLE_STRIP, patch->m_VBSides->m_Index, (GLsizei)patch->m_VBSides->m_Count);
+		deviceCommandContext->Draw(patch->m_VBSides->m_Index, patch->m_VBSides->m_Count);
 
 		// bump stats
 		g_Renderer.m_Stats.m_DrawCalls++;
-		g_Renderer.m_Stats.m_TerrainTris += patch->m_VBSides->m_Count - 2;
+		g_Renderer.m_Stats.m_TerrainTris += patch->m_VBSides->m_Count / 3;
 	}
-
-	CVertexBuffer::Unbind();
-
-	glEnable(GL_CULL_FACE);
 }
 
 void CPatchRData::RenderPriorities(CTextRenderer& textRenderer)
@@ -1220,17 +1311,17 @@ void CPatchRData::BuildWater()
 
 	// Build data for water
 	std::vector<SWaterVertex> water_vertex_data;
-	std::vector<GLushort> water_indices;
+	std::vector<u16> water_indices;
 	u16 water_index_map[PATCH_SIZE+1][PATCH_SIZE+1];
 	memset(water_index_map, 0xFF, sizeof(water_index_map));
 
 	// Build data for shore
 	std::vector<SWaterVertex> water_vertex_data_shore;
-	std::vector<GLushort> water_indices_shore;
+	std::vector<u16> water_indices_shore;
 	u16 water_shore_index_map[PATCH_SIZE+1][PATCH_SIZE+1];
 	memset(water_shore_index_map, 0xFF, sizeof(water_shore_index_map));
 
-	WaterManager* WaterMgr = g_Renderer.GetWaterManager();
+	const WaterManager& waterManager = g_Renderer.GetSceneRenderer().GetWaterManager();
 
 	CPatch* patch = m_Patch;
 	CTerrain* terrain = patch->m_Parent;
@@ -1301,7 +1392,7 @@ void CPatchRData::BuildWater()
 
 				m_WaterBounds += vertex.m_Position;
 
-				vertex.m_WaterData = CVector2D(WaterMgr->m_WindStrength[xx + zz*mapSize], depth);
+				vertex.m_WaterData = CVector2D(waterManager.m_WindStrength[xx + zz*mapSize], depth);
 
 				water_index_map[z+moves[i][1]][x+moves[i][0]] = static_cast<u16>(water_vertex_data.size());
 				water_vertex_data.push_back(vertex);
@@ -1363,81 +1454,104 @@ void CPatchRData::BuildWater()
 	// No vertex buffers if no data generated
 	if (!water_indices.empty())
 	{
-		m_VBWater = g_VBMan.AllocateChunk(sizeof(SWaterVertex), water_vertex_data.size(), GL_STATIC_DRAW, GL_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::WATER);
+		m_VBWater = g_VBMan.AllocateChunk(
+			sizeof(SWaterVertex), water_vertex_data.size(),
+			Renderer::Backend::IBuffer::Type::VERTEX, false,
+			nullptr, CVertexBufferManager::Group::WATER);
 		m_VBWater->m_Owner->UpdateChunkVertices(m_VBWater.Get(), &water_vertex_data[0]);
 
-		m_VBWaterIndices = g_VBMan.AllocateChunk(sizeof(GLushort), water_indices.size(), GL_STATIC_DRAW, GL_ELEMENT_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::WATER);
+		m_VBWaterIndices = g_VBMan.AllocateChunk(
+			sizeof(u16), water_indices.size(),
+			Renderer::Backend::IBuffer::Type::INDEX, false,
+			nullptr, CVertexBufferManager::Group::WATER);
 		m_VBWaterIndices->m_Owner->UpdateChunkVertices(m_VBWaterIndices.Get(), &water_indices[0]);
 	}
 
 	if (!water_indices_shore.empty())
 	{
-		m_VBWaterShore = g_VBMan.AllocateChunk(sizeof(SWaterVertex), water_vertex_data_shore.size(), GL_STATIC_DRAW, GL_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::WATER);
+		m_VBWaterShore = g_VBMan.AllocateChunk(
+			sizeof(SWaterVertex), water_vertex_data_shore.size(),
+			Renderer::Backend::IBuffer::Type::VERTEX, false,
+			nullptr, CVertexBufferManager::Group::WATER);
 		m_VBWaterShore->m_Owner->UpdateChunkVertices(m_VBWaterShore.Get(), &water_vertex_data_shore[0]);
 
 		// Construct indices buffer
-		m_VBWaterIndicesShore = g_VBMan.AllocateChunk(sizeof(GLushort), water_indices_shore.size(), GL_STATIC_DRAW, GL_ELEMENT_ARRAY_BUFFER, nullptr, CVertexBufferManager::Group::WATER);
+		m_VBWaterIndicesShore = g_VBMan.AllocateChunk(
+			sizeof(u16), water_indices_shore.size(),
+			Renderer::Backend::IBuffer::Type::INDEX, false,
+			nullptr, CVertexBufferManager::Group::WATER);
 		m_VBWaterIndicesShore->m_Owner->UpdateChunkVertices(m_VBWaterIndicesShore.Get(), &water_indices_shore[0]);
 	}
 }
 
-void CPatchRData::RenderWater(CShaderProgramPtr& shader, bool onlyShore, bool fixedPipeline)
+void CPatchRData::RenderWaterSurface(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext,
+	const bool bindWaterData)
 {
-	ASSERT(m_UpdateFlags==0);
+	ASSERT(m_UpdateFlags == 0);
 
-	if (g_Renderer.m_SkipSubmit || (!m_VBWater && !m_VBWaterShore))
+	if (!m_VBWater)
 		return;
 
-#if !CONFIG2_GLES
-	if (g_Renderer.GetWaterRenderMode() == WIREFRAME)
-		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
-#endif
+	m_VBWater->m_Owner->UploadIfNeeded(deviceCommandContext);
+	m_VBWaterIndices->m_Owner->UploadIfNeeded(deviceCommandContext);
 
-	if (m_VBWater && !onlyShore)
+	const uint32_t stride = sizeof(SWaterVertex);
+	const uint32_t firstVertexOffset = m_VBWater->m_Index * stride;
+
+	deviceCommandContext->SetVertexAttributeFormat(
+		Renderer::Backend::VertexAttributeStream::POSITION,
+		Renderer::Backend::Format::R32G32B32_SFLOAT,
+		firstVertexOffset + offsetof(SWaterVertex, m_Position), stride,
+		Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+	if (bindWaterData)
 	{
-		SWaterVertex *base=(SWaterVertex *)m_VBWater->m_Owner->Bind();
-
-		// setup data pointers
-		GLsizei stride = sizeof(SWaterVertex);
-		shader->VertexPointer(3, GL_FLOAT, stride, &base[m_VBWater->m_Index].m_Position);
-		if (!fixedPipeline)
-			shader->VertexAttribPointer(str_a_waterInfo, 2, GL_FLOAT, false, stride, &base[m_VBWater->m_Index].m_WaterData);
-
-		shader->AssertPointersBound();
-
-		u8* indexBase = m_VBWaterIndices->m_Owner->Bind();
-		glDrawElements(GL_TRIANGLES, (GLsizei) m_VBWaterIndices->m_Count,
-					   GL_UNSIGNED_SHORT, indexBase + sizeof(u16)*(m_VBWaterIndices->m_Index));
-
-		g_Renderer.m_Stats.m_DrawCalls++;
-		g_Renderer.m_Stats.m_WaterTris += m_VBWaterIndices->m_Count / 3;
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::UV1,
+			Renderer::Backend::Format::R32G32_SFLOAT,
+			firstVertexOffset + offsetof(SWaterVertex, m_WaterData), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
 	}
 
-	if (m_VBWaterShore &&
-	    g_Renderer.GetWaterManager()->m_WaterEffects &&
-	    g_Renderer.GetWaterManager()->m_WaterFancyEffects)
-	{
-		SWaterVertex *base=(SWaterVertex *)m_VBWaterShore->m_Owner->Bind();
+	deviceCommandContext->SetVertexBuffer(0, m_VBWater->m_Owner->GetBuffer());
+	deviceCommandContext->SetIndexBuffer(m_VBWaterIndices->m_Owner->GetBuffer());
 
-		GLsizei stride = sizeof(SWaterVertex);
-		shader->VertexPointer(3, GL_FLOAT, stride, &base[m_VBWaterShore->m_Index].m_Position);
-		if (!fixedPipeline)
-			shader->VertexAttribPointer(str_a_waterInfo, 2, GL_FLOAT, false, stride, &base[m_VBWaterShore->m_Index].m_WaterData);
+	deviceCommandContext->DrawIndexed(m_VBWaterIndices->m_Index, m_VBWaterIndices->m_Count, 0);
 
-		shader->AssertPointersBound();
+	g_Renderer.m_Stats.m_DrawCalls++;
+	g_Renderer.m_Stats.m_WaterTris += m_VBWaterIndices->m_Count / 3;
+}
 
-		u8* indexBase = m_VBWaterIndicesShore->m_Owner->Bind();
-		glDrawElements(GL_TRIANGLES, (GLsizei) m_VBWaterIndicesShore->m_Count,
-					   GL_UNSIGNED_SHORT, indexBase + sizeof(u16)*(m_VBWaterIndicesShore->m_Index));
+void CPatchRData::RenderWaterShore(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext)
+{
+	ASSERT(m_UpdateFlags == 0);
 
-		g_Renderer.m_Stats.m_DrawCalls++;
-		g_Renderer.m_Stats.m_WaterTris += m_VBWaterIndicesShore->m_Count / 3;
-	}
+	if (!m_VBWaterShore)
+		return;
 
-	CVertexBuffer::Unbind();
+	m_VBWaterShore->m_Owner->UploadIfNeeded(deviceCommandContext);
+	m_VBWaterIndicesShore->m_Owner->UploadIfNeeded(deviceCommandContext);
 
-#if !CONFIG2_GLES
-	if (g_Renderer.GetWaterRenderMode() == WIREFRAME)
-		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
-#endif
+	const uint32_t stride = sizeof(SWaterVertex);
+	const uint32_t firstVertexOffset = m_VBWaterShore->m_Index * stride;
+
+	deviceCommandContext->SetVertexAttributeFormat(
+		Renderer::Backend::VertexAttributeStream::POSITION,
+		Renderer::Backend::Format::R32G32B32_SFLOAT,
+		firstVertexOffset + offsetof(SWaterVertex, m_Position), stride,
+		Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+	deviceCommandContext->SetVertexAttributeFormat(
+		Renderer::Backend::VertexAttributeStream::UV1,
+		Renderer::Backend::Format::R32G32_SFLOAT,
+		firstVertexOffset + offsetof(SWaterVertex, m_WaterData), stride,
+		Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+
+	deviceCommandContext->SetVertexBuffer(0, m_VBWaterShore->m_Owner->GetBuffer());
+	deviceCommandContext->SetIndexBuffer(m_VBWaterIndicesShore->m_Owner->GetBuffer());
+
+	deviceCommandContext->DrawIndexed(m_VBWaterIndicesShore->m_Index, m_VBWaterIndicesShore->m_Count, 0);
+
+	g_Renderer.m_Stats.m_DrawCalls++;
+	g_Renderer.m_Stats.m_WaterTris += m_VBWaterIndicesShore->m_Count / 3;
 }
