@@ -1,4 +1,4 @@
-/* Copyright (C) 2021 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -15,10 +15,6 @@
  * along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-/*
- * Water settings (speed, height) and texture management
- */
-
 #include "precompiled.h"
 
 #include "graphics/Terrain.h"
@@ -27,20 +23,23 @@
 #include "graphics/ShaderProgram.h"
 #include "lib/bits.h"
 #include "lib/timer.h"
-#include "lib/tex/tex.h"
-#include "lib/res/graphics/ogl_tex.h"
 #include "maths/MathUtil.h"
 #include "maths/Vector2D.h"
 #include "ps/CLogger.h"
+#include "ps/CStrInternStatic.h"
 #include "ps/Game.h"
+#include "ps/VideoMode.h"
 #include "ps/World.h"
-#include "renderer/WaterManager.h"
+#include "renderer/backend/IDevice.h"
 #include "renderer/Renderer.h"
 #include "renderer/RenderingOptions.h"
+#include "renderer/SceneRenderer.h"
+#include "renderer/WaterManager.h"
 #include "simulation2/Simulation2.h"
 #include "simulation2/components/ICmpWaterManager.h"
 #include "simulation2/components/ICmpRangeManager.h"
 
+#include <algorithm>
 
 struct CoastalPoint
 {
@@ -49,7 +48,8 @@ struct CoastalPoint
 	CVector2D position;
 };
 
-struct SWavesVertex {
+struct SWavesVertex
+{
 	// vertex position
 	CVector3D m_BasePosition;
 	CVector3D m_ApexPosition;
@@ -66,7 +66,7 @@ cassert(sizeof(SWavesVertex) == 64);
 
 struct WaveObject
 {
-	CVertexBuffer::VBChunk* m_VBvertices;
+	CVertexBufferManager::Handle m_VBVertices;
 	CBoundingBoxAligned m_AABB;
 	size_t m_Width;
 	float m_TimeDiff;
@@ -78,16 +78,8 @@ WaterManager::WaterManager()
 	m_RenderWater = false; // disabled until textures are successfully loaded
 	m_WaterHeight = 5.0f;
 
-	m_WaterCurrentTex = 0;
-
-	m_ReflectionTexture = 0;
-	m_RefractionTexture = 0;
 	m_RefTextureSize = 0;
 
-	m_ReflectionFbo = 0;
-	m_RefractionFbo = 0;
-	m_FancyEffectsFBO = 0;
-
 	m_WaterTexTimer = 0.0;
 
 	m_WindAngle = 0.0f;
@@ -97,12 +89,6 @@ WaterManager::WaterManager()
 	m_Murkiness = 0.45f;
 	m_RepeatPeriod = 16.0f;
 
-	m_DistanceHeightmap = NULL;
-	m_BlurredNormalMap = NULL;
-	m_WindStrength = NULL;
-
-	m_ShoreWaves_VBIndices = NULL;
-
 	m_WaterEffects = true;
 	m_WaterFancyEffects = false;
 	m_WaterRealDepth = false;
@@ -113,11 +99,6 @@ WaterManager::WaterManager()
 	m_NeedsReloading = false;
 	m_NeedInfoUpdate = true;
 
-	m_FancyTexture = 0;
-	m_FancyTextureDepth = 0;
-	m_ReflFboDepthTexture = 0;
-	m_RefrFboDepthTexture = 0;
-
 	m_MapSize = 0;
 
 	m_updatei0 = 0;
@@ -131,31 +112,20 @@ WaterManager::~WaterManager()
 	// Cleanup if the caller messed up
 	UnloadWaterTextures();
 
-	for (WaveObject* const& obj : m_ShoreWaves)
-	{
-		if (obj->m_VBvertices)
-			g_VBMan.Release(obj->m_VBvertices);
-		delete obj;
-	}
+	m_ShoreWaves.clear();
+	m_ShoreWavesVBIndices.Reset();
 
-	if (m_ShoreWaves_VBIndices)
-		g_VBMan.Release(m_ShoreWaves_VBIndices);
+	m_DistanceHeightmap.reset();
+	m_WindStrength.reset();
 
-	delete[] m_DistanceHeightmap;
-	delete[] m_BlurredNormalMap;
-	delete[] m_WindStrength;
+	m_FancyEffectsFramebuffer.reset();
+	m_RefractionFramebuffer.reset();
+	m_ReflectionFramebuffer.reset();
 
-	if (!g_Renderer.GetCapabilities().m_PrettyWater)
-		return;
-
-	glDeleteTextures(1, &m_FancyTexture);
-	glDeleteTextures(1, &m_FancyTextureDepth);
-	glDeleteTextures(1, &m_ReflFboDepthTexture);
-	glDeleteTextures(1, &m_RefrFboDepthTexture);
-
-	pglDeleteFramebuffersEXT(1, &m_FancyEffectsFBO);
-	pglDeleteFramebuffersEXT(1, &m_RefractionFbo);
-	pglDeleteFramebuffersEXT(1, &m_ReflectionFbo);
+	m_FancyTexture.reset();
+	m_FancyTextureDepth.reset();
+	m_ReflFboDepthTexture.reset();
+	m_RefrFboDepthTexture.reset();
 }
 
 
@@ -173,30 +143,24 @@ int WaterManager::LoadWaterTextures()
 	{
 		swprintf_s(pathname, ARRAY_SIZE(pathname), L"art/textures/animated/water/default/diffuse%02d.dds", (int)i+1);
 		CTextureProperties textureProps(pathname);
-		textureProps.SetWrap(GL_REPEAT);
+		textureProps.SetAddressMode(
+			Renderer::Backend::Sampler::AddressMode::REPEAT);
 
 		CTexturePtr texture = g_Renderer.GetTextureManager().CreateTexture(textureProps);
 		texture->Prefetch();
 		m_WaterTexture[i] = texture;
 	}
 
-	if (!g_Renderer.GetCapabilities().m_PrettyWater)
-	{
-		// Enable rendering, now that we've succeeded this far
-		m_RenderWater = true;
-		return 0;
-	}
+	m_RenderWater = true;
 
-#if CONFIG2_GLES
-#warning Fix WaterManager::LoadWaterTextures on GLES
-#else
 	// Load normalmaps (for fancy water)
 	ReloadWaterNormalTextures();
 
 	// Load CoastalWaves
 	{
 		CTextureProperties textureProps(L"art/textures/terrain/types/water/coastalWave.png");
-		textureProps.SetWrap(GL_REPEAT);
+		textureProps.SetAddressMode(
+			Renderer::Backend::Sampler::AddressMode::REPEAT);
 		CTexturePtr texture = g_Renderer.GetTextureManager().CreateTexture(textureProps);
 		texture->Prefetch();
 		m_WaveTex = texture;
@@ -205,143 +169,127 @@ int WaterManager::LoadWaterTextures()
 	// Load Foam
 	{
 		CTextureProperties textureProps(L"art/textures/terrain/types/water/foam.png");
-		textureProps.SetWrap(GL_REPEAT);
+		textureProps.SetAddressMode(
+			Renderer::Backend::Sampler::AddressMode::REPEAT);
 		CTexturePtr texture = g_Renderer.GetTextureManager().CreateTexture(textureProps);
 		texture->Prefetch();
 		m_FoamTex = texture;
 	}
 
-	// Use screen-sized textures for minimum artifacts.
-	m_RefTextureSize = g_Renderer.GetHeight();
+	RecreateOrLoadTexturesIfNeeded();
 
-	m_RefTextureSize = round_up_to_pow2(m_RefTextureSize);
-
-	// Create reflection texture
-	glGenTextures(1, &m_ReflectionTexture);
-	glBindTexture(GL_TEXTURE_2D, m_ReflectionTexture);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
-	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, (GLsizei)m_RefTextureSize, (GLsizei)m_RefTextureSize, 0,  GL_RGBA, GL_UNSIGNED_BYTE, 0);
-
-	// Create refraction texture
-	glGenTextures(1, &m_RefractionTexture);
-	glBindTexture(GL_TEXTURE_2D, m_RefractionTexture);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
-	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, (GLsizei)m_RefTextureSize, (GLsizei)m_RefTextureSize, 0,  GL_RGBA, GL_UNSIGNED_BYTE, 0);
-
-	// Create depth textures
-	glGenTextures(1, &m_ReflFboDepthTexture);
-	glBindTexture(GL_TEXTURE_2D, m_ReflFboDepthTexture);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT32, (GLsizei)m_RefTextureSize, (GLsizei)m_RefTextureSize, 0,  GL_DEPTH_COMPONENT, GL_UNSIGNED_SHORT, NULL);
-
-	glGenTextures(1, &m_RefrFboDepthTexture);
-	glBindTexture(GL_TEXTURE_2D, m_RefrFboDepthTexture);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT32, (GLsizei)m_RefTextureSize, (GLsizei)m_RefTextureSize, 0,  GL_DEPTH_COMPONENT, GL_UNSIGNED_SHORT, NULL);
-
-	// Create the Fancy Effects texture
-	glGenTextures(1, &m_FancyTexture);
-	glBindTexture(GL_TEXTURE_2D, m_FancyTexture);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-
-	glGenTextures(1, &m_FancyTextureDepth);
-	glBindTexture(GL_TEXTURE_2D, m_FancyTextureDepth);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
-
-	glBindTexture(GL_TEXTURE_2D, 0);
-
-	Resize();
-
-	// Create the water framebuffers
-
-	GLint currentFbo;
-	glGetIntegerv(GL_FRAMEBUFFER_BINDING_EXT, &currentFbo);
-
-	m_ReflectionFbo = 0;
-	pglGenFramebuffersEXT(1, &m_ReflectionFbo);
-	pglBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_ReflectionFbo);
-	pglFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_ReflectionTexture, 0);
-	pglFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, m_ReflFboDepthTexture, 0);
-
-	ogl_WarnIfError();
-
-	GLenum status = pglCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
-	if (status != GL_FRAMEBUFFER_COMPLETE_EXT)
-	{
-		LOGWARNING("Reflection framebuffer object incomplete: 0x%04X", status);
-		g_RenderingOptions.SetWaterReflection(false);
-		UpdateQuality();
-	}
-
-	m_RefractionFbo = 0;
-	pglGenFramebuffersEXT(1, &m_RefractionFbo);
-	pglBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_RefractionFbo);
-	pglFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_RefractionTexture, 0);
-	pglFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, m_RefrFboDepthTexture, 0);
-
-	ogl_WarnIfError();
-
-	status = pglCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
-	if (status != GL_FRAMEBUFFER_COMPLETE_EXT)
-	{
-		LOGWARNING("Refraction framebuffer object incomplete: 0x%04X", status);
-		g_RenderingOptions.SetWaterRefraction(false);
-		UpdateQuality();
-	}
-
-	pglGenFramebuffersEXT(1, &m_FancyEffectsFBO);
-	pglBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_FancyEffectsFBO);
-	pglFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_FancyTexture, 0);
-	pglFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, m_FancyTextureDepth, 0);
-
-	ogl_WarnIfError();
-
-	status = pglCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
-	if (status != GL_FRAMEBUFFER_COMPLETE_EXT)
-	{
-		LOGWARNING("Fancy Effects framebuffer object incomplete: 0x%04X", status);
-		g_RenderingOptions.SetWaterRefraction(false);
-		UpdateQuality();
-	}
-
-	pglBindFramebufferEXT(GL_FRAMEBUFFER_EXT, currentFbo);
-
-	// Enable rendering, now that we've succeeded this far
-	m_RenderWater = true;
-#endif
 	return 0;
 }
 
-
-///////////////////////////////////////////////////////////////////
-// Resize: Updates the fancy water textures.
-void WaterManager::Resize()
+void WaterManager::RecreateOrLoadTexturesIfNeeded()
 {
-	glBindTexture(GL_TEXTURE_2D, m_FancyTexture);
-	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, (GLsizei)g_Renderer.GetWidth(), (GLsizei)g_Renderer.GetHeight(), 0,  GL_RGBA, GL_UNSIGNED_SHORT, NULL);
+	Renderer::Backend::IDevice* backendDevice = g_VideoMode.GetBackendDevice();
 
-	glBindTexture(GL_TEXTURE_2D, m_FancyTextureDepth);
-	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT32, (GLsizei)g_Renderer.GetWidth(), (GLsizei)g_Renderer.GetHeight(), 0,  GL_DEPTH_COMPONENT, GL_UNSIGNED_SHORT, NULL);
+	// Use screen-sized textures for minimum artifacts.
+	const size_t newRefTextureSize = round_up_to_pow2(g_Renderer.GetHeight());
 
-	glBindTexture(GL_TEXTURE_2D, 0);
+	if (m_RefTextureSize != newRefTextureSize)
+	{
+		m_ReflectionFramebuffer.reset();
+		m_ReflectionTexture.reset();
+		m_ReflFboDepthTexture.reset();
+
+		m_RefractionFramebuffer.reset();
+		m_RefractionTexture.reset();
+		m_RefrFboDepthTexture.reset();
+
+		m_RefTextureSize = newRefTextureSize;
+	}
+
+	// Create reflection textures.
+	const bool needsReflectionTextures =
+		g_RenderingOptions.GetWaterEffects() &&
+		g_RenderingOptions.GetWaterReflection();
+	if (needsReflectionTextures && !m_ReflectionTexture)
+	{
+		m_ReflectionTexture = backendDevice->CreateTexture2D("WaterReflectionTexture",
+			Renderer::Backend::Format::R8G8B8A8_UNORM, m_RefTextureSize, m_RefTextureSize,
+			Renderer::Backend::Sampler::MakeDefaultSampler(
+				Renderer::Backend::Sampler::Filter::LINEAR,
+				Renderer::Backend::Sampler::AddressMode::MIRRORED_REPEAT));
+
+		m_ReflFboDepthTexture = backendDevice->CreateTexture2D("WaterReflectionDepthTexture",
+			Renderer::Backend::Format::D32, m_RefTextureSize, m_RefTextureSize,
+			Renderer::Backend::Sampler::MakeDefaultSampler(
+				Renderer::Backend::Sampler::Filter::NEAREST,
+				Renderer::Backend::Sampler::AddressMode::REPEAT));
+
+		m_ReflectionFramebuffer = backendDevice->CreateFramebuffer("ReflectionFramebuffer",
+			m_ReflectionTexture.get(), m_ReflFboDepthTexture.get(), CColor(0.5f, 0.5f, 1.0f, 0.0f));
+		if (!m_ReflectionFramebuffer)
+		{
+			g_RenderingOptions.SetWaterReflection(false);
+			UpdateQuality();
+		}
+	}
+
+	// Create refraction textures.
+	const bool needsRefractionTextures =
+		g_RenderingOptions.GetWaterEffects() &&
+		g_RenderingOptions.GetWaterRefraction();
+	if (needsRefractionTextures && !m_RefractionTexture)
+	{
+		m_RefractionTexture = backendDevice->CreateTexture2D("WaterRefractionTexture",
+			Renderer::Backend::Format::R8G8B8A8_UNORM, m_RefTextureSize, m_RefTextureSize,
+			Renderer::Backend::Sampler::MakeDefaultSampler(
+				Renderer::Backend::Sampler::Filter::LINEAR,
+				Renderer::Backend::Sampler::AddressMode::MIRRORED_REPEAT));
+
+		m_RefrFboDepthTexture = backendDevice->CreateTexture2D("WaterRefractionDepthTexture",
+			Renderer::Backend::Format::D32, m_RefTextureSize, m_RefTextureSize,
+			Renderer::Backend::Sampler::MakeDefaultSampler(
+				Renderer::Backend::Sampler::Filter::NEAREST,
+				Renderer::Backend::Sampler::AddressMode::REPEAT));
+
+		m_RefractionFramebuffer = backendDevice->CreateFramebuffer("RefractionFramebuffer",
+			m_RefractionTexture.get(), m_RefrFboDepthTexture.get(), CColor(1.0f, 0.0f, 0.0f, 0.0f));
+		if (!m_RefractionFramebuffer)
+		{
+			g_RenderingOptions.SetWaterRefraction(false);
+			UpdateQuality();
+		}
+	}
+
+	const uint32_t newWidth = static_cast<uint32_t>(g_Renderer.GetWidth());
+	const uint32_t newHeight = static_cast<uint32_t>(g_Renderer.GetHeight());
+	if (m_FancyTexture && (m_FancyTexture->GetWidth() != newWidth || m_FancyTexture->GetHeight() != newHeight))
+	{
+		m_FancyEffectsFramebuffer.reset();
+		m_FancyTexture.reset();
+		m_FancyTextureDepth.reset();
+	}
+
+	// Create the Fancy Effects textures.
+	const bool needsFancyTextures =
+		g_RenderingOptions.GetWaterEffects() &&
+		g_RenderingOptions.GetWaterFancyEffects();
+	if (needsFancyTextures && !m_FancyTexture)
+	{
+		m_FancyTexture = backendDevice->CreateTexture2D("WaterFancyTexture",
+			Renderer::Backend::Format::R8G8B8A8_UNORM, g_Renderer.GetWidth(), g_Renderer.GetHeight(),
+			Renderer::Backend::Sampler::MakeDefaultSampler(
+				Renderer::Backend::Sampler::Filter::LINEAR,
+				Renderer::Backend::Sampler::AddressMode::REPEAT));
+
+		m_FancyTextureDepth = backendDevice->CreateTexture2D("WaterFancyDepthTexture",
+			Renderer::Backend::Format::D32, g_Renderer.GetWidth(), g_Renderer.GetHeight(),
+			Renderer::Backend::Sampler::MakeDefaultSampler(
+				Renderer::Backend::Sampler::Filter::LINEAR,
+				Renderer::Backend::Sampler::AddressMode::REPEAT));
+
+		m_FancyEffectsFramebuffer = backendDevice->CreateFramebuffer("FancyEffectsFramebuffer",
+			m_FancyTexture.get(), m_FancyTextureDepth.get());
+		if (!m_FancyEffectsFramebuffer)
+		{
+			g_RenderingOptions.SetWaterRefraction(false);
+			UpdateQuality();
+		}
+	}
 }
 
 void WaterManager::ReloadWaterNormalTextures()
@@ -351,8 +299,9 @@ void WaterManager::ReloadWaterNormalTextures()
 	{
 		swprintf_s(pathname, ARRAY_SIZE(pathname), L"art/textures/animated/water/%ls/normal00%02d.png", m_WaterType.c_str(), static_cast<int>(i) + 1);
 		CTextureProperties textureProps(pathname);
-		textureProps.SetWrap(GL_REPEAT);
-		textureProps.SetMaxAnisotropy(4);
+		textureProps.SetAddressMode(
+			Renderer::Backend::Sampler::AddressMode::REPEAT);
+		textureProps.SetAnisotropicFilter(true);
 
 		CTexturePtr texture = g_Renderer.GetTextureManager().CreateTexture(textureProps);
 		texture->Prefetch();
@@ -364,19 +313,16 @@ void WaterManager::ReloadWaterNormalTextures()
 // Unload water textures
 void WaterManager::UnloadWaterTextures()
 {
-	for(size_t i = 0; i < ARRAY_SIZE(m_WaterTexture); i++)
+	for (size_t i = 0; i < ARRAY_SIZE(m_WaterTexture); i++)
 		m_WaterTexture[i].reset();
 
-	if (!g_Renderer.GetCapabilities().m_PrettyWater)
-		return;
-
-	for(size_t i = 0; i < ARRAY_SIZE(m_NormalMap); i++)
+	for (size_t i = 0; i < ARRAY_SIZE(m_NormalMap); i++)
 		m_NormalMap[i].reset();
 
-	glDeleteTextures(1, &m_ReflectionTexture);
-	glDeleteTextures(1, &m_RefractionTexture);
-	pglDeleteFramebuffersEXT(1, &m_RefractionFbo);
-	pglDeleteFramebuffersEXT(1, &m_ReflectionFbo);
+	m_RefractionFramebuffer.reset();
+	m_ReflectionFramebuffer.reset();
+	m_ReflectionTexture.reset();
+	m_RefractionTexture.reset();
 }
 
 template<bool Transpose>
@@ -437,10 +383,10 @@ void WaterManager::RecomputeDistanceHeightmap()
 	// we want to look ahead some distance, but not too much (less efficient and not interesting). This is our lookahead.
 	const size_t maxLevel = 5;
 
-	if (m_DistanceHeightmap == NULL)
+	if (!m_DistanceHeightmap)
 	{
-		m_DistanceHeightmap = new float[SideSize*SideSize];
-		std::fill(m_DistanceHeightmap, m_DistanceHeightmap + SideSize*SideSize, (float)maxLevel);
+		m_DistanceHeightmap = std::make_unique<float[]>(SideSize * SideSize);
+		std::fill(m_DistanceHeightmap.get(), m_DistanceHeightmap.get() + SideSize * SideSize, static_cast<float>(maxLevel));
 	}
 
 	// Create a manhattan-distance heightmap.
@@ -448,8 +394,8 @@ void WaterManager::RecomputeDistanceHeightmap()
 
 	u16* heightmap = terrain->GetHeightMap();
 
-	ComputeDirection<false>(m_DistanceHeightmap, heightmap, m_WaterHeight, SideSize, maxLevel);
-	ComputeDirection<true>(m_DistanceHeightmap, heightmap, m_WaterHeight, SideSize, maxLevel);
+	ComputeDirection<false>(m_DistanceHeightmap.get(), heightmap, m_WaterHeight, SideSize, maxLevel);
+	ComputeDirection<true>(m_DistanceHeightmap.get(), heightmap, m_WaterHeight, SideSize, maxLevel);
 }
 
 // This requires m_DistanceHeightmap to be defined properly.
@@ -462,19 +408,8 @@ void WaterManager::CreateWaveMeshes()
 	if (!terrain || !terrain->GetHeightMap())
 		return;
 
-	for (WaveObject* const& obj : m_ShoreWaves)
-	{
-		if (obj->m_VBvertices)
-			g_VBMan.Release(obj->m_VBvertices);
-		delete obj;
-	}
 	m_ShoreWaves.clear();
-
-	if (m_ShoreWaves_VBIndices)
-	{
-		g_VBMan.Release(m_ShoreWaves_VBIndices);
-		m_ShoreWaves_VBIndices = NULL;
-	}
+	m_ShoreWavesVBIndices.Reset();
 
 	if (m_Waviness < 5.0f && m_WaterType != L"ocean")
 		return;
@@ -600,13 +535,13 @@ void WaterManager::CreateWaveMeshes()
 	}
 
 	// Fourth step: create waves themselves, using those chains. We basically create subchains.
-	GLushort waveSizes = 14;	// maximal size in width.
+	u16 waveSizes = 14;	// maximal size in width.
 
 	// Construct indices buffer (we can afford one for all of them)
-	std::vector<GLushort> water_indices;
-	for (GLushort a = 0; a < waveSizes - 1; ++a)
+	std::vector<u16> water_indices;
+	for (u16 a = 0; a < waveSizes - 1; ++a)
 	{
-		for (GLushort rect = 0; rect < 7; ++rect)
+		for (u16 rect = 0; rect < 7; ++rect)
 		{
 			water_indices.push_back(a * 9 + rect);
 			water_indices.push_back(a * 9 + 9 + rect);
@@ -617,11 +552,15 @@ void WaterManager::CreateWaveMeshes()
 		}
 	}
 	// Generic indexes, max-length
-	m_ShoreWaves_VBIndices = g_VBMan.Allocate(sizeof(GLushort), water_indices.size(), GL_STATIC_DRAW, GL_ELEMENT_ARRAY_BUFFER);
-	m_ShoreWaves_VBIndices->m_Owner->UpdateChunkVertices(m_ShoreWaves_VBIndices, &water_indices[0]);
+	m_ShoreWavesVBIndices = g_VBMan.AllocateChunk(
+		sizeof(u16), water_indices.size(),
+		Renderer::Backend::IBuffer::Type::INDEX, false,
+		nullptr, CVertexBufferManager::Group::WATER);
+	m_ShoreWavesVBIndices->m_Owner->UpdateChunkVertices(m_ShoreWavesVBIndices.Get(), &water_indices[0]);
 
 	float diff = (rand() % 50) / 5.0f;
 
+	std::vector<SWavesVertex> vertices, reversed;
 	for (size_t i = 0; i < CoastalPointsChains.size(); ++i)
 	{
 		for (size_t j = 0; j < CoastalPointsChains[i].size()-waveSizes; ++j)
@@ -629,14 +568,14 @@ void WaterManager::CreateWaveMeshes()
 			if (CoastalPointsChains[i].size()- 1 - j < waveSizes)
 				break;
 
-			GLushort width = waveSizes;
+			u16 width = waveSizes;
 
 			// First pass to get some parameters out.
 			float outmost = 0.0f;	// how far to move on the shore.
 			float avgDepth = 0.0f;
 			int sign = 1;
 			CVector2D firstPerp(0,0), perp(0,0), lastPerp(0,0);
-			for (GLushort a = 0; a < waveSizes;++a)
+			for (u16 a = 0; a < waveSizes;++a)
 			{
 				lastPerp = perp;
 				perp = CVector2D(0,0);
@@ -709,15 +648,15 @@ void WaterManager::CreateWaveMeshes()
 			}
 			// we passed the checks, we can create a wave of size "width".
 
-			WaveObject* shoreWave = new WaveObject;
-			std::vector<SWavesVertex> vertices;
-			vertices.reserve(9*width);
+			std::unique_ptr<WaveObject> shoreWave = std::make_unique<WaveObject>();
+			vertices.clear();
+			vertices.reserve(9 * width);
 
 			shoreWave->m_Width = width;
 			shoreWave->m_TimeDiff = diff;
 			diff += (rand() % 100) / 25.0f + 4.0f;
 
-			for (GLushort a = 0; a < width;++a)
+			for (u16 a = 0; a < width;++a)
 			{
 				perp = CVector2D(0,0);
 				int nb = 0;
@@ -823,98 +762,123 @@ void WaterManager::CreateWaveMeshes()
 			if (sign == 1)
 			{
 				// Let's do some fancy reversing.
-				std::vector<SWavesVertex> reversed;
+				reversed.clear();
 				reversed.reserve(vertices.size());
-				for (int a = width-1; a >= 0; --a)
+				for (int a = width - 1; a >= 0; --a)
 				{
 					for (size_t t = 0; t < 9; ++t)
-						reversed.push_back(vertices[a*9+t]);
+						reversed.push_back(vertices[a * 9 + t]);
 				}
-				vertices = reversed;
+				std::swap(vertices, reversed);
 			}
 			j += width/2-1;
 
-			shoreWave->m_VBvertices = g_VBMan.Allocate(sizeof(SWavesVertex), vertices.size(), GL_STATIC_DRAW, GL_ARRAY_BUFFER);
-			shoreWave->m_VBvertices->m_Owner->UpdateChunkVertices(shoreWave->m_VBvertices, &vertices[0]);
+			shoreWave->m_VBVertices = g_VBMan.AllocateChunk(
+				sizeof(SWavesVertex), vertices.size(),
+				Renderer::Backend::IBuffer::Type::VERTEX, false,
+				nullptr, CVertexBufferManager::Group::WATER);
+			shoreWave->m_VBVertices->m_Owner->UpdateChunkVertices(shoreWave->m_VBVertices.Get(), &vertices[0]);
 
-			m_ShoreWaves.push_back(shoreWave);
+			m_ShoreWaves.emplace_back(std::move(shoreWave));
 		}
 	}
 }
 
-void WaterManager::RenderWaves(const CFrustum& frustrum)
+void WaterManager::RenderWaves(
+	Renderer::Backend::IDeviceCommandContext* deviceCommandContext,
+	const CFrustum& frustrum)
 {
-#if CONFIG2_GLES
-#warning Fix WaterManager::RenderWaves on GLES
-#else
-	if (g_Renderer.m_SkipSubmit || !m_WaterFancyEffects)
+	GPU_SCOPED_LABEL(deviceCommandContext, "Render Waves");
+	if (!m_WaterFancyEffects)
 		return;
 
-	pglBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_FancyEffectsFBO);
+	deviceCommandContext->SetGraphicsPipelineState(
+		Renderer::Backend::MakeDefaultGraphicsPipelineStateDesc());
+	deviceCommandContext->SetFramebuffer(m_FancyEffectsFramebuffer.get());
+	deviceCommandContext->ClearFramebuffer();
 
-	GLuint attachments[1] = { GL_COLOR_ATTACHMENT0_EXT };
-	pglDrawBuffers(1, attachments);
+	CShaderTechniquePtr tech = g_Renderer.GetShaderManager().LoadEffect(str_water_waves);
+	deviceCommandContext->SetGraphicsPipelineState(
+		tech->GetGraphicsPipelineStateDesc());
+	deviceCommandContext->BeginPass();
+	Renderer::Backend::IShaderProgram* shader = tech->GetShader();
 
-	glClearColor(0.0f,0.0f, 0.0f,0.0f);
-	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
+	m_WaveTex->UploadBackendTextureIfNeeded(deviceCommandContext);
+	m_FoamTex->UploadBackendTextureIfNeeded(deviceCommandContext);
 
-	glEnable(GL_BLEND);
-	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-	glEnable(GL_DEPTH_TEST);
-	glDepthFunc(GL_ALWAYS);
+	deviceCommandContext->SetTexture(
+		shader->GetBindingSlot(str_waveTex), m_WaveTex->GetBackendTexture());
+	deviceCommandContext->SetTexture(
+		shader->GetBindingSlot(str_foamTex), m_FoamTex->GetBackendTexture());
 
-	CShaderDefines none;
-	CShaderProgramPtr shader = g_Renderer.GetShaderManager().LoadProgram("glsl/waves", none);
-
-	shader->Bind();
-
-	shader->BindTexture(str_waveTex, m_WaveTex);
-	shader->BindTexture(str_foamTex, m_FoamTex);
-
-	shader->Uniform(str_time, (float)m_WaterTexTimer);
-	shader->Uniform(str_transform, g_Renderer.GetViewCamera().GetViewProjection());
+	deviceCommandContext->SetUniform(
+		shader->GetBindingSlot(str_time), static_cast<float>(m_WaterTexTimer));
+	const CMatrix3D transform =
+		g_Renderer.GetSceneRenderer().GetViewCamera().GetViewProjection();
+	deviceCommandContext->SetUniform(
+		shader->GetBindingSlot(str_transform), transform.AsFloatArray());
 
 	for (size_t a = 0; a < m_ShoreWaves.size(); ++a)
 	{
 		if (!frustrum.IsBoxVisible(m_ShoreWaves[a]->m_AABB))
 			continue;
 
-		CVertexBuffer::VBChunk* VBchunk = m_ShoreWaves[a]->m_VBvertices;
-		SWavesVertex* base = (SWavesVertex*)VBchunk->m_Owner->Bind();
+		CVertexBuffer::VBChunk* VBchunk = m_ShoreWaves[a]->m_VBVertices.Get();
+		VBchunk->m_Owner->UploadIfNeeded(deviceCommandContext);
+		m_ShoreWavesVBIndices->m_Owner->UploadIfNeeded(deviceCommandContext);
 
-		// setup data pointers
-		GLsizei stride = sizeof(SWavesVertex);
-		shader->VertexPointer(3, GL_FLOAT, stride, &base[VBchunk->m_Index].m_BasePosition);
-		shader->TexCoordPointer(GL_TEXTURE0, 2, GL_UNSIGNED_BYTE, stride, &base[VBchunk->m_Index].m_UV);
-		//	NormalPointer(gl_FLOAT, stride, &base[m_VBWater->m_Index].m_UV)
-		pglVertexAttribPointerARB(2, 2, GL_FLOAT, GL_FALSE, stride, &base[VBchunk->m_Index].m_PerpVect);	// replaces commented above because my normal is vec2
-		shader->VertexAttribPointer(str_a_apexPosition, 3, GL_FLOAT, false, stride, &base[VBchunk->m_Index].m_ApexPosition);
-		shader->VertexAttribPointer(str_a_splashPosition, 3, GL_FLOAT, false, stride, &base[VBchunk->m_Index].m_SplashPosition);
-		shader->VertexAttribPointer(str_a_retreatPosition, 3, GL_FLOAT, false, stride, &base[VBchunk->m_Index].m_RetreatPosition);
+		const uint32_t stride = sizeof(SWavesVertex);
+		const uint32_t firstVertexOffset = VBchunk->m_Index * stride;
 
-		shader->AssertPointersBound();
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::POSITION,
+			Renderer::Backend::Format::R32G32B32_SFLOAT,
+			firstVertexOffset + offsetof(SWavesVertex, m_BasePosition), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::NORMAL,
+			Renderer::Backend::Format::R32G32_SFLOAT,
+			firstVertexOffset + offsetof(SWavesVertex, m_PerpVect), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::UV0,
+			Renderer::Backend::Format::R8G8_UINT,
+			firstVertexOffset + offsetof(SWavesVertex, m_UV), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
 
-		shader->Uniform(str_translation, m_ShoreWaves[a]->m_TimeDiff);
-		shader->Uniform(str_width, (int)m_ShoreWaves[a]->m_Width);
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::UV1,
+			Renderer::Backend::Format::R32G32B32_SFLOAT,
+			firstVertexOffset + offsetof(SWavesVertex, m_ApexPosition), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::UV2,
+			Renderer::Backend::Format::R32G32B32_SFLOAT,
+			firstVertexOffset + offsetof(SWavesVertex, m_SplashPosition), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
+		deviceCommandContext->SetVertexAttributeFormat(
+			Renderer::Backend::VertexAttributeStream::UV3,
+			Renderer::Backend::Format::R32G32B32_SFLOAT,
+			firstVertexOffset + offsetof(SWavesVertex, m_RetreatPosition), stride,
+			Renderer::Backend::VertexAttributeRate::PER_VERTEX, 0);
 
-		u8* indexBase = m_ShoreWaves_VBIndices->m_Owner->Bind();
-		glDrawElements(GL_TRIANGLES, (GLsizei) (m_ShoreWaves[a]->m_Width-1)*(7*6),
-					   GL_UNSIGNED_SHORT, indexBase + sizeof(u16)*(m_ShoreWaves_VBIndices->m_Index));
+		deviceCommandContext->SetUniform(
+			shader->GetBindingSlot(str_translation), m_ShoreWaves[a]->m_TimeDiff);
+		deviceCommandContext->SetUniform(
+			shader->GetBindingSlot(str_width), static_cast<float>(m_ShoreWaves[a]->m_Width));
 
-		shader->Uniform(str_translation, m_ShoreWaves[a]->m_TimeDiff + 6.0f);
+		deviceCommandContext->SetVertexBuffer(0, VBchunk->m_Owner->GetBuffer());
+		deviceCommandContext->SetIndexBuffer(m_ShoreWavesVBIndices->m_Owner->GetBuffer());
 
-		// TODO: figure out why this doesn't work.
-		//g_Renderer.m_Stats.m_DrawCalls++;
-		//g_Renderer.m_Stats.m_WaterTris += m_ShoreWaves_VBIndices->m_Count / 3;
+		const uint32_t indexCount = (m_ShoreWaves[a]->m_Width - 1) * (7 * 6);
+		deviceCommandContext->DrawIndexed(m_ShoreWavesVBIndices->m_Index, indexCount, 0);
 
-		CVertexBuffer::Unbind();
+		g_Renderer.GetStats().m_DrawCalls++;
+		g_Renderer.GetStats().m_WaterTris += indexCount / 3;
 	}
-	shader->Unbind();
-	pglBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
-
-	glDisable(GL_BLEND);
-	glDepthFunc(GL_LEQUAL);
-#endif
+	deviceCommandContext->EndPass();
+	deviceCommandContext->SetFramebuffer(
+		deviceCommandContext->GetDevice()->GetCurrentBackbuffer());
 }
 
 void WaterManager::RecomputeWaterData()
@@ -934,17 +898,18 @@ void WaterManager::RecomputeWindStrength()
 	if (m_MapSize <= 0)
 		return;
 
-	if (m_WindStrength == nullptr)
-		m_WindStrength = new float[m_MapSize*m_MapSize];
+	if (!m_WindStrength)
+		m_WindStrength = std::make_unique<float[]>(m_MapSize * m_MapSize);
 
 	CTerrain* terrain = g_Game->GetWorld()->GetTerrain();
 	if (!terrain || !terrain->GetHeightMap())
 		return;
 
-	CVector2D windDir = CVector2D(cos(m_WindAngle),sin(m_WindAngle));
+	CVector2D windDir = CVector2D(cos(m_WindAngle), sin(m_WindAngle));
 
-	ssize_t windX = round(1.f / windDir.X);
-	ssize_t windY = round(1.f / windDir.Y);
+	int stepSize = 10;
+	ssize_t windX = -round(stepSize * windDir.X);
+	ssize_t windY = -round(stepSize * windDir.Y);
 
 	struct SWindPoint {
 		SWindPoint(size_t x, size_t y, float strength) : X(x), Y(y), windStrength(strength) {}
@@ -959,7 +924,7 @@ void WaterManager::RecomputeWindStrength()
 	// Compute starting points (one or two edges of the map) and how much to move each computation increment.
 	if (fabs(windDir.X) < 0.01f)
 	{
-		movement.emplace_back(0, windY);
+		movement.emplace_back(0, windY > 0.f ? 1 : -1);
 		startingPoints.reserve(m_MapSize);
 		size_t start = windY > 0 ? 0 : m_MapSize - 1;
 		for (size_t x = 0; x < m_MapSize; ++x)
@@ -967,7 +932,8 @@ void WaterManager::RecomputeWindStrength()
 	}
 	else if (fabs(windDir.Y) < 0.01f)
 	{
-		movement.emplace_back(windX, 0);
+		movement.emplace_back(windX > 0.f ? 1 : - 1, 0);
+		startingPoints.reserve(m_MapSize);
 		size_t start = windX > 0 ? 0 : m_MapSize - 1;
 		for (size_t z = 0; z < m_MapSize; ++z)
 			startingPoints.emplace_back(start, z, 0.f);
@@ -1022,7 +988,7 @@ void WaterManager::RecomputeWindStrength()
 				// Adjust speed based on height difference, a positive height difference slowly increases speed (simulate venturi effect)
 				// and a lower height reduces speed (wind protection from hills/...)
 				float heightDiff = std::max(m_WaterHeight, terrain->GetVertexGroundLevel(point.X + movement[step].first, point.Y + movement[step].second)) -
-								   std::max(m_WaterHeight, terrain->GetVertexGroundLevel(point.X, point.Y));
+					std::max(m_WaterHeight, terrain->GetVertexGroundLevel(point.X, point.Y));
 				if (heightDiff > 0.f)
 					point.windStrength = std::min(2.f, point.windStrength + std::min(4.f, heightDiff) / 40.f);
 				else
@@ -1054,9 +1020,8 @@ void WaterManager::SetMapSize(size_t size)
 	m_updatej0 = 0;
 	m_updatej1 = size;
 
-	SAFE_ARRAY_DELETE(m_DistanceHeightmap);
-	SAFE_ARRAY_DELETE(m_BlurredNormalMap);
-	SAFE_ARRAY_DELETE(m_WindStrength);
+	m_DistanceHeightmap.reset();
+	m_WindStrength.reset();
 }
 
 ////////////////////////////////////////////////////////////////////////
@@ -1090,7 +1055,21 @@ void WaterManager::UpdateQuality()
 	}
 }
 
-bool WaterManager::WillRenderFancyWater()
+bool WaterManager::WillRenderFancyWater() const
 {
-	return m_RenderWater && g_RenderingOptions.GetWaterEffects() && g_Renderer.GetCapabilities().m_PrettyWater;
+	return
+		m_RenderWater && g_VideoMode.GetBackend() != CVideoMode::Backend::GL_ARB &&
+		g_RenderingOptions.GetWaterEffects();
+}
+
+size_t WaterManager::GetCurrentTextureIndex(const double& period) const
+{
+	ENSURE(period > 0.0);
+	return static_cast<size_t>(m_WaterTexTimer * ARRAY_SIZE(m_WaterTexture) / period) % ARRAY_SIZE(m_WaterTexture);
+}
+
+size_t WaterManager::GetNextTextureIndex(const double& period) const
+{
+	ENSURE(period > 0.0);
+	return (GetCurrentTextureIndex(period) + 1) % ARRAY_SIZE(m_WaterTexture);
 }
