@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2021 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -36,6 +36,17 @@
 
 #include "js/TypeDecls.h"
 
+// Complete with a few additional ones.
+namespace JS
+{
+class HandleValueArray;
+}
+
+namespace js {
+class BaseProxyHandler;
+}
+
+class ScriptContext;
 class ScriptInterface;
 class ScriptRequest;
 
