@@ -1,4 +1,4 @@
-/* Copyright (C) 2021 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -19,8 +19,11 @@
 
 #include "Simulation2.h"
 
+#include "scriptinterface/FunctionWrapper.h"
 #include "scriptinterface/ScriptContext.h"
 #include "scriptinterface/ScriptInterface.h"
+#include "scriptinterface/JSON.h"
+#include "scriptinterface/StructuredClone.h"
 
 #include "simulation2/MessageTypes.h"
 #include "simulation2/system/ComponentManager.h"
@@ -44,15 +47,16 @@
 #include "ps/Util.h"
 #include "ps/XML/Xeromyces.h"
 
+#include <fstream>
 #include <iomanip>
+#include <memory>
 
 class CSimulation2Impl
 {
 public:
-	CSimulation2Impl(CUnitManager* unitManager, shared_ptr<ScriptContext> cx, CTerrain* terrain) :
+	CSimulation2Impl(CUnitManager* unitManager, std::shared_ptr<ScriptContext> cx, CTerrain* terrain) :
 		m_SimContext(), m_ComponentManager(m_SimContext, cx),
 		m_EnableOOSLog(false), m_EnableSerializationTest(false), m_RejoinTestTurn(-1), m_TestingRejoin(false),
-		m_SecondaryTerrain(nullptr), m_SecondaryContext(nullptr), m_SecondaryComponentManager(nullptr), m_SecondaryLoadedScripts(nullptr),
 		m_MapSettings(cx->GetGeneralJSContext()), m_InitAttributes(cx->GetGeneralJSContext())
 	{
 		m_SimContext.m_UnitManager = unitManager;
@@ -80,11 +84,6 @@ public:
 
 	~CSimulation2Impl()
 	{
-		delete m_SecondaryTerrain;
-		delete m_SecondaryContext;
-		delete m_SecondaryComponentManager;
-		delete m_SecondaryLoadedScripts;
-
 		UnregisterFileReloadFunc(ReloadChangedFileCB, this);
 	}
 
@@ -142,11 +141,11 @@ public:
 	int m_RejoinTestTurn;
 	bool m_TestingRejoin;
 
-	// Secondary simulation
-	CTerrain* m_SecondaryTerrain;
-	CSimContext* m_SecondaryContext;
-	CComponentManager* m_SecondaryComponentManager;
-	std::set<VfsPath>* m_SecondaryLoadedScripts;
+	// Secondary simulation (NB: order matters for destruction).
+	std::unique_ptr<CComponentManager> m_SecondaryComponentManager;
+	std::unique_ptr<CTerrain> m_SecondaryTerrain;
+	std::unique_ptr<CSimContext> m_SecondaryContext;
+	std::unique_ptr<std::set<VfsPath>> m_SecondaryLoadedScripts;
 
 	struct SerializationTestState
 	{
@@ -164,7 +163,7 @@ public:
 	void InitRNGSeedSimulation();
 	void InitRNGSeedAI();
 
-	static std::vector<SimulationCommand> CloneCommandsFromOtherCompartment(const ScriptInterface& oldScript, const ScriptInterface& newScript,
+	static std::vector<SimulationCommand> CloneCommandsFromOtherCompartment(const ScriptInterface& newScript, const ScriptInterface& oldScript,
 		const std::vector<SimulationCommand>& commands)
 	{
 		std::vector<SimulationCommand> newCommands;
@@ -173,8 +172,8 @@ public:
 		ScriptRequest rqNew(newScript);
 		for (const SimulationCommand& command : commands)
 		{
-			JS::RootedValue tmpCommand(rqNew.cx, newScript.CloneValueFromOtherCompartment(oldScript, command.data));
-			newScript.FreezeObject(tmpCommand, true);
+			JS::RootedValue tmpCommand(rqNew.cx, Script::CloneValueFromOtherCompartment(newScript, oldScript, command.data));
+			Script::FreezeObject(rqNew, tmpCommand, true);
 			SimulationCommand cmd(command.player, rqNew.cx, tmpCommand);
 			newCommands.emplace_back(std::move(cmd));
 		}
@@ -212,10 +211,11 @@ bool CSimulation2Impl::LoadScripts(CComponentManager& componentManager, std::set
 bool CSimulation2Impl::LoadTriggerScripts(CComponentManager& componentManager, JS::HandleValue mapSettings, std::set<VfsPath>* loadedScripts)
 {
 	bool ok = true;
-	if (componentManager.GetScriptInterface().HasProperty(mapSettings, "TriggerScripts"))
+	ScriptRequest rq(componentManager.GetScriptInterface());
+	if (Script::HasProperty(rq, mapSettings, "TriggerScripts"))
 	{
 		std::vector<std::string> scriptNames;
-		componentManager.GetScriptInterface().GetProperty(mapSettings, "TriggerScripts", scriptNames);
+		Script::GetProperty(rq, mapSettings, "TriggerScripts", scriptNames);
 		for (const std::string& triggerScript : scriptNames)
 		{
 			std::string scriptName = "maps/" + triggerScript;
@@ -337,8 +337,9 @@ void CSimulation2Impl::ReportSerializationFailure(
 void CSimulation2Impl::InitRNGSeedSimulation()
 {
 	u32 seed = 0;
-	if (!m_ComponentManager.GetScriptInterface().HasProperty(m_MapSettings, "Seed") ||
-	    !m_ComponentManager.GetScriptInterface().GetProperty(m_MapSettings, "Seed", seed))
+	ScriptRequest rq(m_ComponentManager.GetScriptInterface());
+	if (!Script::HasProperty(rq, m_MapSettings, "Seed") ||
+		!Script::GetProperty(rq, m_MapSettings, "Seed", seed))
 		LOGWARNING("CSimulation2Impl::InitRNGSeedSimulation: No seed value specified - using %d", seed);
 
 	m_ComponentManager.SetRNGSeed(seed);
@@ -347,8 +348,9 @@ void CSimulation2Impl::InitRNGSeedSimulation()
 void CSimulation2Impl::InitRNGSeedAI()
 {
 	u32 seed = 0;
-	if (!m_ComponentManager.GetScriptInterface().HasProperty(m_MapSettings, "AISeed") ||
-	    !m_ComponentManager.GetScriptInterface().GetProperty(m_MapSettings, "AISeed", seed))
+	ScriptRequest rq(m_ComponentManager.GetScriptInterface());
+	if (!Script::HasProperty(rq, m_MapSettings, "AISeed") ||
+		!Script::GetProperty(rq, m_MapSettings, "AISeed", seed))
 		LOGWARNING("CSimulation2Impl::InitRNGSeedAI: No seed value specified - using %d", seed);
 
 	CmpPtr<ICmpAIManager> cmpAIManager(m_SimContext, SYSTEM_ENTITY);
@@ -403,28 +405,25 @@ void CSimulation2Impl::Update(int turnLength, const std::vector<SimulationComman
 		if (startRejoinTest)
 			debug_printf("Initializing the secondary simulation\n");
 
-		delete m_SecondaryTerrain;
-		m_SecondaryTerrain = new CTerrain();
+		m_SecondaryTerrain = std::make_unique<CTerrain>();
 
-		delete m_SecondaryContext;
-		m_SecondaryContext = new CSimContext();
-		m_SecondaryContext->m_Terrain = m_SecondaryTerrain;
+		m_SecondaryContext = std::make_unique<CSimContext>();
+		m_SecondaryContext->m_Terrain = m_SecondaryTerrain.get();
 
-		delete m_SecondaryComponentManager;
-		m_SecondaryComponentManager = new CComponentManager(*m_SecondaryContext, scriptInterface.GetContext());
+		m_SecondaryComponentManager = std::make_unique<CComponentManager>(*m_SecondaryContext, scriptInterface.GetContext());
 		m_SecondaryComponentManager->LoadComponentTypes();
 
-		delete m_SecondaryLoadedScripts;
-		m_SecondaryLoadedScripts = new std::set<VfsPath>();
-		ENSURE(LoadDefaultScripts(*m_SecondaryComponentManager, m_SecondaryLoadedScripts));
+		m_SecondaryLoadedScripts = std::make_unique<std::set<VfsPath>>();
+		ENSURE(LoadDefaultScripts(*m_SecondaryComponentManager, m_SecondaryLoadedScripts.get()));
 		ResetComponentState(*m_SecondaryComponentManager, false, false);
 
+		ScriptRequest rq(scriptInterface);
+
 		// Load the trigger scripts after we have loaded the simulation.
 		{
 			ScriptRequest rq2(m_SecondaryComponentManager->GetScriptInterface());
-			JS::RootedValue mapSettingsCloned(rq2.cx,
-				m_SecondaryComponentManager->GetScriptInterface().CloneValueFromOtherCompartment(scriptInterface, m_MapSettings));
-			ENSURE(LoadTriggerScripts(*m_SecondaryComponentManager, mapSettingsCloned, m_SecondaryLoadedScripts));
+			JS::RootedValue mapSettingsCloned(rq2.cx, Script::CloneValueFromOtherCompartment(m_SecondaryComponentManager->GetScriptInterface(), scriptInterface, m_MapSettings));
+			ENSURE(LoadTriggerScripts(*m_SecondaryComponentManager, mapSettingsCloned, m_SecondaryLoadedScripts.get()));
 		}
 
 		// Load the map into the secondary simulation
@@ -433,7 +432,7 @@ void CSimulation2Impl::Update(int turnLength, const std::vector<SimulationComman
 		std::unique_ptr<CMapReader> mapReader = std::make_unique<CMapReader>();
 
 		std::string mapType;
-		scriptInterface.GetProperty(m_InitAttributes, "mapType", mapType);
+		Script::GetProperty(rq, m_InitAttributes, "mapType", mapType);
 		if (mapType == "random")
 		{
 			// TODO: support random map scripts
@@ -442,12 +441,12 @@ void CSimulation2Impl::Update(int turnLength, const std::vector<SimulationComman
 		else
 		{
 			std::wstring mapFile;
-			scriptInterface.GetProperty(m_InitAttributes, "map", mapFile);
+			Script::GetProperty(rq, m_InitAttributes, "map", mapFile);
 
 			VfsPath mapfilename = VfsPath(mapFile).ChangeExtension(L".pmp");
 			mapReader->LoadMap(mapfilename, *scriptInterface.GetContext(), JS::UndefinedHandleValue,
-				m_SecondaryTerrain, NULL, NULL, NULL, NULL, NULL, NULL,
-				NULL, NULL, m_SecondaryContext, INVALID_PLAYER, true); // throws exception on failure
+				m_SecondaryTerrain.get(), NULL, NULL, NULL, NULL, NULL, NULL,
+				NULL, NULL, m_SecondaryContext.get(), INVALID_PLAYER, true); // throws exception on failure
 		}
 
 		LDR_EndRegistering();
@@ -476,7 +475,7 @@ void CSimulation2Impl::Update(int turnLength, const std::vector<SimulationComman
 			ENSURE(m_ComponentManager.ComputeStateHash(primaryStateAfter.hash, false));
 
 		UpdateComponents(*m_SecondaryContext, turnLengthFixed,
-			CloneCommandsFromOtherCompartment(scriptInterface, m_SecondaryComponentManager->GetScriptInterface(), commands));
+			CloneCommandsFromOtherCompartment(m_SecondaryComponentManager->GetScriptInterface(), scriptInterface, commands));
 		SerializationTestState secondaryStateAfter;
 		ENSURE(m_SecondaryComponentManager->SerializeState(secondaryStateAfter.state));
 		if (serializationTestHash)
@@ -511,11 +510,6 @@ void CSimulation2Impl::Update(int turnLength, const std::vector<SimulationComman
 	if (m_EnableOOSLog)
 		DumpState();
 
-	// Start computing AI for the next turn
-	CmpPtr<ICmpAIManager> cmpAIManager(m_SimContext, SYSTEM_ENTITY);
-	if (cmpAIManager)
-		cmpAIManager->StartComputation();
-
 	++m_TurnNumber;
 }
 
@@ -528,7 +522,7 @@ void CSimulation2Impl::UpdateComponents(CSimContext& simContext, fixed turnLengt
 
 	CmpPtr<ICmpPathfinder> cmpPathfinder(simContext, SYSTEM_ENTITY);
 	if (cmpPathfinder)
-		cmpPathfinder->FetchAsyncResultsAndSendMessages();
+		cmpPathfinder->SendRequestedPaths();
 
 	{
 		PROFILE2("Sim - Update Start");
@@ -536,10 +530,6 @@ void CSimulation2Impl::UpdateComponents(CSimContext& simContext, fixed turnLengt
 		componentManager.BroadcastMessage(msgTurnStart);
 	}
 
-	// Push AI commands onto the queue before we use them
-	CmpPtr<ICmpAIManager> cmpAIManager(simContext, SYSTEM_ENTITY);
-	if (cmpAIManager)
-		cmpAIManager->PushCommands();
 
 	CmpPtr<ICmpCommandQueue> cmpCommandQueue(simContext, SYSTEM_ENTITY);
 	if (cmpCommandQueue)
@@ -549,7 +539,7 @@ void CSimulation2Impl::UpdateComponents(CSimContext& simContext, fixed turnLengt
 	if (cmpPathfinder)
 	{
 		cmpPathfinder->StartProcessingMoves(true);
-		cmpPathfinder->FetchAsyncResultsAndSendMessages();
+		cmpPathfinder->SendRequestedPaths();
 	}
 	// Send all the update phases
 	{
@@ -567,7 +557,7 @@ void CSimulation2Impl::UpdateComponents(CSimContext& simContext, fixed turnLengt
 	if (cmpPathfinder)
 	{
 		cmpPathfinder->StartProcessingMoves(true);
-		cmpPathfinder->FetchAsyncResultsAndSendMessages();
+		cmpPathfinder->SendRequestedPaths();
 	}
 
 	{
@@ -584,6 +574,14 @@ void CSimulation2Impl::UpdateComponents(CSimContext& simContext, fixed turnLengt
 	// Clean up any entities destroyed during the simulation update
 	componentManager.FlushDestroyedComponents();
 
+	// Compute AI immediately at turn's end.
+	CmpPtr<ICmpAIManager> cmpAIManager(simContext, SYSTEM_ENTITY);
+	if (cmpAIManager)
+	{
+		cmpAIManager->StartComputation();
+		cmpAIManager->PushCommands();
+	}
+
 	// Process all remaining moves
 	if (cmpPathfinder)
 	{
@@ -637,7 +635,7 @@ void CSimulation2Impl::DumpState()
 
 ////////////////////////////////////////////////////////////////
 
-CSimulation2::CSimulation2(CUnitManager* unitManager, shared_ptr<ScriptContext> cx, CTerrain* terrain) :
+CSimulation2::CSimulation2(CUnitManager* unitManager, std::shared_ptr<ScriptContext> cx, CTerrain* terrain) :
 	m(new CSimulation2Impl(unitManager, cx, terrain))
 {
 }
@@ -734,7 +732,7 @@ void CSimulation2::PreInitGame()
 {
 	ScriptRequest rq(GetScriptInterface());
 	JS::RootedValue global(rq.cx, rq.globalValue());
-	GetScriptInterface().CallFunctionVoid(global, "PreInitGame");
+	ScriptFunction::CallVoid(rq, global, "PreInitGame");
 }
 
 void CSimulation2::InitGame()
@@ -744,9 +742,9 @@ void CSimulation2::InitGame()
 
 	JS::RootedValue settings(rq.cx);
 	JS::RootedValue tmpInitAttributes(rq.cx, GetInitAttributes());
-	GetScriptInterface().GetProperty(tmpInitAttributes, "settings", &settings);
+	Script::GetProperty(rq, tmpInitAttributes, "settings", &settings);
 
-	GetScriptInterface().CallFunctionVoid(global, "InitGame", settings);
+	ScriptFunction::CallVoid(rq, global, "InitGame", settings);
 }
 
 void CSimulation2::Update(int turnLength)
@@ -815,7 +813,7 @@ void CSimulation2::GetInitAttributes(JS::MutableHandleValue ret)
 
 void CSimulation2::SetMapSettings(const std::string& settings)
 {
-	m->m_ComponentManager.GetScriptInterface().ParseJSON(settings, &m->m_MapSettings);
+	Script::ParseJSON(ScriptRequest(m->m_ComponentManager.GetScriptInterface()), settings, &m->m_MapSettings);
 }
 
 void CSimulation2::SetMapSettings(JS::HandleValue settings)
@@ -828,7 +826,7 @@ void CSimulation2::SetMapSettings(JS::HandleValue settings)
 
 std::string CSimulation2::GetMapSettingsString()
 {
-	return m->m_ComponentManager.GetScriptInterface().StringifyJSON(&m->m_MapSettings);
+	return Script::StringifyJSON(ScriptRequest(m->m_ComponentManager.GetScriptInterface()), &m->m_MapSettings);
 }
 
 void CSimulation2::GetMapSettings(JS::MutableHandleValue ret)
@@ -840,7 +838,7 @@ void CSimulation2::LoadPlayerSettings(bool newPlayers)
 {
 	ScriptRequest rq(GetScriptInterface());
 	JS::RootedValue global(rq.cx, rq.globalValue());
-	GetScriptInterface().CallFunctionVoid(global, "LoadPlayerSettings", m->m_MapSettings, newPlayers);
+	ScriptFunction::CallVoid(rq, global, "LoadPlayerSettings", m->m_MapSettings, newPlayers);
 }
 
 void CSimulation2::LoadMapSettings()
@@ -850,9 +848,9 @@ void CSimulation2::LoadMapSettings()
 	JS::RootedValue global(rq.cx, rq.globalValue());
 
 	// Initialize here instead of in Update()
-	GetScriptInterface().CallFunctionVoid(global, "LoadMapSettings", m->m_MapSettings);
+	ScriptFunction::CallVoid(rq, global, "LoadMapSettings", m->m_MapSettings);
 
-	GetScriptInterface().FreezeObject(m->m_InitAttributes, true);
+	Script::FreezeObject(rq, m->m_InitAttributes, true);
 	GetScriptInterface().SetGlobal("InitAttributes", m->m_InitAttributes, true, true, true);
 
 	if (!m->m_StartupScript.empty())
@@ -884,6 +882,7 @@ bool CSimulation2::ComputeStateHash(std::string& outHash, bool quick)
 
 bool CSimulation2::DumpDebugState(std::ostream& stream)
 {
+	stream << "sim turn: " << m->m_TurnNumber << std::endl;
 	return m->m_ComponentManager.DumpDebugState(stream, true);
 }
 
@@ -946,11 +945,6 @@ std::vector<std::string> CSimulation2::GetRMSData()
 	return GetJSONData(L"maps/random/");
 }
 
-std::vector<std::string> CSimulation2::GetCivData()
-{
-	return GetJSONData(L"simulation/data/civs/");
-}
-
 std::vector<std::string> CSimulation2::GetVictoryConditiondData()
 {
 	return GetJSONData(L"simulation/data/settings/victory_conditions/");
@@ -995,8 +989,8 @@ std::string CSimulation2::GetAIData()
 	// Build single JSON string with array of AI data
 	JS::RootedValue ais(rq.cx);
 
-	if (!ScriptInterface::CreateObject(rq, &ais, "AIData", aiData))
+	if (!Script::CreateObject(rq, &ais, "AIData", aiData))
 		return std::string();
 
-	return scriptInterface.StringifyJSON(&ais);
+	return Script::StringifyJSON(rq, &ais);
 }
