@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -32,7 +32,10 @@
 #include "ps/scripting/JSInterface_VFS.h"
 #include "ps/TemplateLoader.h"
 #include "ps/Util.h"
+#include "scriptinterface/FunctionWrapper.h"
 #include "scriptinterface/ScriptContext.h"
+#include "scriptinterface/StructuredClone.h"
+#include "scriptinterface/JSON.h"
 #include "simulation2/components/ICmpAIInterface.h"
 #include "simulation2/components/ICmpCommandQueue.h"
 #include "simulation2/components/ICmpObstructionManager.h"
@@ -54,27 +57,24 @@ extern void QuitEngine();
  * AI is primarily scripted, and the CCmpAIManager component defined here
  * takes care of managing all the scripts.
  *
- * To avoid slow AI scripts causing jerky rendering, they are run in a background
- * thread (maintained by CAIWorker) so that it's okay if they take a whole simulation
- * turn before returning their results (though preferably they shouldn't use nearly
- * that much CPU).
+ * The original idea was to run CAIWorker in a separate thread to prevent
+ * slow AIs from impacting framerate. However, copying the game-state every turn
+ * proved difficult and rather slow itself (and isn't threadable, obviously).
+ * For these reasons, the design was changed to a single-thread, same-compartment, different-realm design.
+ * The AI can therefore directly use the simulation data via the 'Sim' & 'SimEngine' globals.
+ * As a result, a lof of the code is still designed to be "thread-ready", but this no longer matters.
  *
- * CCmpAIManager grabs the world state after each turn (making use of AIInterface.js
- * and AIProxy.js to decide what data to include) then passes it to CAIWorker.
- * The AI scripts will then run asynchronously and return a list of commands to execute.
- * Any attempts to read the command list (including indirectly via serialization)
- * will block until it's actually completed, so the rest of the engine should avoid
- * reading it for as long as possible.
+ * TODO: despite the above, it would still be useful to allow the AI to run tasks asynchronously (and off-thread).
+ * This could be implemented by having a separate JS runtime in a different thread,
+ * that runs tasks and returns after a distinct # of simulation turns (to maintain determinism).
  *
- * JS::Values are passed between the game and AI threads using ScriptInterface::StructuredClone.
- *
- * TODO: actually the thread isn't implemented yet, because performance hasn't been
- * sufficiently problematic to justify the complexity yet, but the CAIWorker interface
- * is designed to hopefully support threading when we want it.
+ * Note also that the RL Interface, by default, uses the 'AI representation'.
+ * This representation, alimented by the JS AIInterface/AIProxy tandem, is likely to grow smaller over time
+ * as the AI uses more sim data directly.
  */
 
 /**
- * Implements worker thread for CCmpAIManager.
+ * AI computation orchestator for CCmpAIManager.
  */
 class CAIWorker
 {
@@ -84,7 +84,7 @@ private:
 		NONCOPYABLE(CAIPlayer);
 	public:
 		CAIPlayer(CAIWorker& worker, const std::wstring& aiName, player_id_t player, u8 difficulty, const std::wstring& behavior,
-				shared_ptr<ScriptInterface> scriptInterface) :
+			std::shared_ptr<ScriptInterface> scriptInterface) :
 			m_Worker(worker), m_AIName(aiName), m_Player(player), m_Difficulty(difficulty), m_Behavior(behavior),
 			m_ScriptInterface(scriptInterface), m_Obj(scriptInterface->GetGeneralJSContext())
 		{
@@ -113,39 +113,39 @@ private:
 			JS::RootedValue objectWithConstructor(rq.cx); // object that should contain the constructor function
 			JS::RootedValue global(rq.cx, rq.globalValue());
 			JS::RootedValue ctor(rq.cx);
-			if (!m_ScriptInterface->HasProperty(metadata, "moduleName"))
+			if (!Script::HasProperty(rq, metadata, "moduleName"))
 			{
 				LOGERROR("Failed to create AI player: %s: missing 'moduleName'", path.string8());
 				return false;
 			}
 
-			m_ScriptInterface->GetProperty(metadata, "moduleName", moduleName);
-			if (!m_ScriptInterface->GetProperty(global, moduleName.c_str(), &objectWithConstructor)
+			Script::GetProperty(rq, metadata, "moduleName", moduleName);
+			if (!Script::GetProperty(rq, global, moduleName.c_str(), &objectWithConstructor)
 			    || objectWithConstructor.isUndefined())
 			{
 				LOGERROR("Failed to create AI player: %s: can't find the module that should contain the constructor: '%s'", path.string8(), moduleName);
 				return false;
 			}
 
-			if (!m_ScriptInterface->GetProperty(metadata, "constructor", constructor))
+			if (!Script::GetProperty(rq, metadata, "constructor", constructor))
 			{
 				LOGERROR("Failed to create AI player: %s: missing 'constructor'", path.string8());
 				return false;
 			}
 
 			// Get the constructor function from the loaded scripts
-			if (!m_ScriptInterface->GetProperty(objectWithConstructor, constructor.c_str(), &ctor)
+			if (!Script::GetProperty(rq, objectWithConstructor, constructor.c_str(), &ctor)
 			    || ctor.isNull())
 			{
 				LOGERROR("Failed to create AI player: %s: can't find constructor '%s'", path.string8(), constructor);
 				return false;
 			}
 
-			m_ScriptInterface->GetProperty(metadata, "useShared", m_UseSharedComponent);
+			Script::GetProperty(rq, metadata, "useShared", m_UseSharedComponent);
 
 			// Set up the data to pass as the constructor argument
 			JS::RootedValue settings(rq.cx);
-			ScriptInterface::CreateObject(
+			Script::CreateObject(
 				rq,
 				&settings,
 				"player", m_Player,
@@ -155,7 +155,7 @@ private:
 			if (!m_UseSharedComponent)
 			{
 				ENSURE(m_Worker.m_HasLoadedEntityTemplates);
-				m_ScriptInterface->SetProperty(settings, "templates", m_Worker.m_EntityTemplates, false);
+				Script::SetProperty(rq, settings, "templates", m_Worker.m_EntityTemplates, false);
 			}
 
 			JS::RootedValueVector argv(rq.cx);
@@ -173,19 +173,22 @@ private:
 		void Run(JS::HandleValue state, int playerID)
 		{
 			m_Commands.clear();
-			m_ScriptInterface->CallFunctionVoid(m_Obj, "HandleMessage", state, playerID);
+			ScriptRequest rq(m_ScriptInterface);
+			ScriptFunction::CallVoid(rq, m_Obj, "HandleMessage", state, playerID);
 		}
 		// overloaded with a sharedAI part.
 		// javascript can handle both natively on the same function.
 		void Run(JS::HandleValue state, int playerID, JS::HandleValue SharedAI)
 		{
 			m_Commands.clear();
-			m_ScriptInterface->CallFunctionVoid(m_Obj, "HandleMessage", state, playerID, SharedAI);
+			ScriptRequest rq(m_ScriptInterface);
+			ScriptFunction::CallVoid(rq, m_Obj, "HandleMessage", state, playerID, SharedAI);
 		}
 		void InitAI(JS::HandleValue state, JS::HandleValue SharedAI)
 		{
 			m_Commands.clear();
-			m_ScriptInterface->CallFunctionVoid(m_Obj, "Init", state, m_Player, SharedAI);
+			ScriptRequest rq(m_ScriptInterface);
+			ScriptFunction::CallVoid(rq, m_Obj, "Init", state, m_Player, SharedAI);
 		}
 
 		CAIWorker& m_Worker;
@@ -197,30 +200,46 @@ private:
 
 		// Take care to keep this declaration before heap rooted members. Destructors of heap rooted
 		// members have to be called before the context destructor.
-		shared_ptr<ScriptInterface> m_ScriptInterface;
+		std::shared_ptr<ScriptInterface> m_ScriptInterface;
 
 		JS::PersistentRootedValue m_Obj;
-		std::vector<ScriptInterface::StructuredClone > m_Commands;
+		std::vector<Script::StructuredClone> m_Commands;
 	};
 
 public:
 	struct SCommandSets
 	{
 		player_id_t player;
-		std::vector<ScriptInterface::StructuredClone > commands;
+		std::vector<Script::StructuredClone> commands;
 	};
 
 	CAIWorker() :
-		m_ScriptInterface(new ScriptInterface("Engine", "AI", g_ScriptContext)),
 		m_TurnNum(0),
 		m_CommandsComputed(true),
 		m_HasLoadedEntityTemplates(false),
-		m_HasSharedComponent(false),
-		m_EntityTemplates(g_ScriptContext->GetGeneralJSContext()),
-		m_SharedAIObj(g_ScriptContext->GetGeneralJSContext()),
-		m_PassabilityMapVal(g_ScriptContext->GetGeneralJSContext()),
-		m_TerritoryMapVal(g_ScriptContext->GetGeneralJSContext())
+		m_HasSharedComponent(false)
 	{
+	}
+
+	~CAIWorker()
+	{
+		// Init will always be called.
+		JS_RemoveExtraGCRootsTracer(m_ScriptInterface->GetGeneralJSContext(), Trace, this);
+	}
+
+	void Init(const ScriptInterface& simInterface)
+	{
+		// Create the script interface in the same compartment as the simulation interface.
+		// This will allow us to directly share data from the sim to the AI (and vice versa, should the need arise).
+		m_ScriptInterface = std::make_shared<ScriptInterface>("Engine", "AI", simInterface);
+
+		ScriptRequest rq(m_ScriptInterface);
+
+		m_EntityTemplates.init(rq.cx);
+		m_SharedAIObj.init(rq.cx);
+		m_PassabilityMapVal.init(rq.cx);
+		m_TerritoryMapVal.init(rq.cx);
+
 
 		m_ScriptInterface->ReplaceNondeterministicRNG(m_RNG);
 
@@ -228,24 +247,34 @@ public:
 
 		JS_AddExtraGCRootsTracer(m_ScriptInterface->GetGeneralJSContext(), Trace, this);
 
-		m_ScriptInterface->RegisterFunction<void, int, JS::HandleValue, CAIWorker::PostCommand>("PostCommand");
-		m_ScriptInterface->RegisterFunction<void, std::wstring, CAIWorker::IncludeModule>("IncludeModule");
-		m_ScriptInterface->RegisterFunction<void, CAIWorker::ExitProgram>("Exit");
+		{
+			ScriptRequest simrq(simInterface);
+			// Register the sim globals for easy & explicit access. Mark it replaceable for hotloading.
+			JS::RootedValue global(rq.cx, simrq.globalValue());
+			m_ScriptInterface->SetGlobal("Sim", global, true);
+			JS::RootedValue scope(rq.cx, JS::ObjectValue(*simrq.nativeScope.get()));
+			m_ScriptInterface->SetGlobal("SimEngine", scope, true);
+		}
 
-		m_ScriptInterface->RegisterFunction<JS::Value, JS::HandleValue, JS::HandleValue, pass_class_t, CAIWorker::ComputePath>("ComputePath");
+#define REGISTER_FUNC_NAME(func, name) \
+	ScriptFunction::Register<&CAIWorker::func, ScriptInterface::ObjectFromCBData<CAIWorker>>(rq, name);
 
-		m_ScriptInterface->RegisterFunction<void, std::wstring, std::vector<u32>, u32, u32, u32, CAIWorker::DumpImage>("DumpImage");
-		m_ScriptInterface->RegisterFunction<CParamNode, std::string, CAIWorker::GetTemplate>("GetTemplate");
+		REGISTER_FUNC_NAME(PostCommand, "PostCommand");
+		REGISTER_FUNC_NAME(LoadScripts, "IncludeModule");
+		ScriptFunction::Register<QuitEngine>(rq, "Exit");
 
-		JSI_VFS::RegisterScriptFunctions_Simulation(*(m_ScriptInterface.get()));
+		REGISTER_FUNC_NAME(ComputePathScript, "ComputePath");
+
+		REGISTER_FUNC_NAME(DumpImage, "DumpImage");
+		REGISTER_FUNC_NAME(GetTemplate, "GetTemplate");
+
+#undef REGISTER_FUNC_NAME
+
+		JSI_VFS::RegisterScriptFunctions_ReadOnlySimulation(rq);
 
 		// Globalscripts may use VFS script functions
 		m_ScriptInterface->LoadGlobalScripts();
-	}
 
-	~CAIWorker()
-	{
-		JS_RemoveExtraGCRootsTracer(m_ScriptInterface->GetGeneralJSContext(), Trace, this);
 	}
 
 	bool HasLoadedEntityTemplates() const { return m_HasLoadedEntityTemplates; }
@@ -279,27 +308,14 @@ public:
 		return true;
 	}
 
-	static void IncludeModule(ScriptInterface::CmptPrivate* pCmptPrivate, const std::wstring& name)
-	{
-		ENSURE(pCmptPrivate->pCBData);
-		CAIWorker* self = static_cast<CAIWorker*> (pCmptPrivate->pCBData);
-		self->LoadScripts(name);
-	}
-
-	static void PostCommand(ScriptInterface::CmptPrivate* pCmptPrivate, int playerid, JS::HandleValue cmd)
-	{
-		ENSURE(pCmptPrivate->pCBData);
-		CAIWorker* self = static_cast<CAIWorker*> (pCmptPrivate->pCBData);
-		self->PostCommand(playerid, cmd);
-	}
-
 	void PostCommand(int playerid, JS::HandleValue cmd)
 	{
+		ScriptRequest rq(m_ScriptInterface);
 		for (size_t i=0; i<m_Players.size(); i++)
 		{
 			if (m_Players[i]->m_Player == playerid)
 			{
-				m_Players[i]->m_Commands.push_back(m_ScriptInterface->WriteStructuredClone(cmd));
+				m_Players[i]->m_Commands.push_back(Script::WriteStructuredClone(rq, cmd));
 				return;
 			}
 		}
@@ -307,22 +323,19 @@ public:
 		LOGERROR("Invalid playerid in PostCommand!");
 	}
 
-	static JS::Value ComputePath(ScriptInterface::CmptPrivate* pCmptPrivate,
-		JS::HandleValue position, JS::HandleValue goal, pass_class_t passClass)
+	JS::Value ComputePathScript(JS::HandleValue position, JS::HandleValue goal, pass_class_t passClass)
 	{
-		ENSURE(pCmptPrivate->pCBData);
-		CAIWorker* self = static_cast<CAIWorker*> (pCmptPrivate->pCBData);
-		ScriptRequest rq(self->m_ScriptInterface);
+		ScriptRequest rq(m_ScriptInterface);
 
 		CFixedVector2D pos, goalPos;
 		std::vector<CFixedVector2D> waypoints;
 		JS::RootedValue retVal(rq.cx);
 
-		self->m_ScriptInterface->FromJSVal<CFixedVector2D>(rq, position, pos);
-		self->m_ScriptInterface->FromJSVal<CFixedVector2D>(rq, goal, goalPos);
+		Script::FromJSVal(rq, position, pos);
+		Script::FromJSVal(rq, goal, goalPos);
 
-		self->ComputePath(pos, goalPos, passClass, waypoints);
-		self->m_ScriptInterface->ToJSVal<std::vector<CFixedVector2D> >(rq, &retVal, waypoints);
+		ComputePath(pos, goalPos, passClass, waypoints);
+		Script::ToJSVal(rq, &retVal, waypoints);
 
 		return retVal;
 	}
@@ -337,14 +350,6 @@ public:
 			waypoints.emplace_back(wp.x, wp.z);
 	}
 
-	static CParamNode GetTemplate(ScriptInterface::CmptPrivate* pCmptPrivate, const std::string& name)
-	{
-		ENSURE(pCmptPrivate->pCBData);
-		CAIWorker* self = static_cast<CAIWorker*> (pCmptPrivate->pCBData);
-
-		return self->GetTemplate(name);
-	}
-
 	CParamNode GetTemplate(const std::string& name)
 	{
 		if (!m_TemplateLoader.TemplateExists(name))
@@ -352,15 +357,10 @@ public:
 		return m_TemplateLoader.GetTemplateFileData(name).GetChild("Entity");
 	}
 
-	static void ExitProgram(ScriptInterface::CmptPrivate* UNUSED(pCmptPrivate))
-	{
-		QuitEngine();
-	}
-
 	/**
 	 * Debug function for AI scripts to dump 2D array data (e.g. terrain tile weights).
 	 */
-	static void DumpImage(ScriptInterface::CmptPrivate* UNUSED(pCmptPrivate), const std::wstring& name, const std::vector<u32>& data, u32 w, u32 h, u32 max)
+	void DumpImage(const std::wstring& name, const std::vector<u32>& data, u32 w, u32 h, u32 max)
 	{
 		// TODO: this is totally not threadsafe.
 		VfsPath filename = L"screenshots/aidump/" + name;
@@ -382,7 +382,7 @@ public:
 
 		const size_t img_size = w * h * bpp/8;
 		const size_t hdr_size = tex_hdr_size(filename);
-		shared_ptr<u8> buf;
+		std::shared_ptr<u8> buf;
 		AllocateAligned(buf, hdr_size+img_size, maxSectorSize);
 		Tex t;
 		if (t.wrap(w, h, bpp, flags, buf, hdr_size) < 0)
@@ -424,13 +424,13 @@ public:
 		JS::RootedValue AIModule(rq.cx);
 		JS::RootedValue global(rq.cx, rq.globalValue());
 		JS::RootedValue ctor(rq.cx);
-		if (!m_ScriptInterface->GetProperty(global, "API3", &AIModule) || AIModule.isUndefined())
+		if (!Script::GetProperty(rq, global, "API3", &AIModule) || AIModule.isUndefined())
 		{
 			LOGERROR("Failed to create shared AI component: %s: can't find module '%s'", path.string8(), "API3");
 			return false;
 		}
 
-		if (!m_ScriptInterface->GetProperty(AIModule, "SharedScript", &ctor)
+		if (!Script::GetProperty(rq, AIModule, "SharedScript", &ctor)
 		    || ctor.isUndefined())
 		{
 			LOGERROR("Failed to create shared AI component: %s: can't find constructor '%s'", path.string8(), "SharedScript");
@@ -439,19 +439,19 @@ public:
 
 		// Set up the data to pass as the constructor argument
 		JS::RootedValue playersID(rq.cx);
-		ScriptInterface::CreateObject(rq, &playersID);
+		Script::CreateObject(rq, &playersID);
 
 		for (size_t i = 0; i < m_Players.size(); ++i)
 		{
 			JS::RootedValue val(rq.cx);
-			m_ScriptInterface->ToJSVal(rq, &val, m_Players[i]->m_Player);
-			m_ScriptInterface->SetPropertyInt(playersID, i, val, true);
+			Script::ToJSVal(rq, &val, m_Players[i]->m_Player);
+			Script::SetPropertyInt(rq, playersID, i, val, true);
 		}
 
 		ENSURE(m_HasLoadedEntityTemplates);
 
 		JS::RootedValue settings(rq.cx);
-		ScriptInterface::CreateObject(
+		Script::CreateObject(
 			rq,
 			&settings,
 			"players", playersID,
@@ -472,7 +472,7 @@ public:
 
 	bool AddPlayer(const std::wstring& aiName, player_id_t player, u8 difficulty, const std::wstring& behavior)
 	{
-		shared_ptr<CAIPlayer> ai = std::make_shared<CAIPlayer>(*this, aiName, player, difficulty, behavior, m_ScriptInterface);
+		std::shared_ptr<CAIPlayer> ai = std::make_shared<CAIPlayer>(*this, aiName, player, difficulty, behavior, m_ScriptInterface);
 		if (!ai->Initialise())
 			return false;
 
@@ -485,7 +485,7 @@ public:
 		return true;
 	}
 
-	bool RunGamestateInit(const ScriptInterface::StructuredClone& gameState, const Grid<NavcellData>& passabilityMap, const Grid<u8>& territoryMap,
+	bool RunGamestateInit(const Script::StructuredClone& gameState, const Grid<NavcellData>& passabilityMap, const Grid<u8>& territoryMap,
 		const std::map<std::string, pass_class_t>& nonPathfindingPassClassMasks, const std::map<std::string, pass_class_t>& pathfindingPassClassMasks)
 	{
 		// this will be run last by InitGame.js, passing the full game representation.
@@ -494,9 +494,9 @@ public:
 		ScriptRequest rq(m_ScriptInterface);
 
 		JS::RootedValue state(rq.cx);
-		m_ScriptInterface->ReadStructuredClone(gameState, &state);
-		ScriptInterface::ToJSVal(rq, &m_PassabilityMapVal, passabilityMap);
-		ScriptInterface::ToJSVal(rq, &m_TerritoryMapVal, territoryMap);
+		Script::ReadStructuredClone(rq, gameState, &state);
+		Script::ToJSVal(rq, &m_PassabilityMapVal, passabilityMap);
+		Script::ToJSVal(rq, &m_TerritoryMapVal, territoryMap);
 
 		m_PassabilityMap = passabilityMap;
 		m_NonPathfindingPassClasses = nonPathfindingPassClassMasks;
@@ -507,9 +507,9 @@ public:
 
 		if (m_HasSharedComponent)
 		{
-			m_ScriptInterface->SetProperty(state, "passabilityMap", m_PassabilityMapVal, true);
-			m_ScriptInterface->SetProperty(state, "territoryMap", m_TerritoryMapVal, true);
-			m_ScriptInterface->CallFunctionVoid(m_SharedAIObj, "init", state);
+			Script::SetProperty(rq, state, "passabilityMap", m_PassabilityMapVal, true);
+			Script::SetProperty(rq, state, "territoryMap", m_TerritoryMapVal, true);
+			ScriptFunction::CallVoid(rq, m_SharedAIObj, "init", state);
 
 			for (size_t i = 0; i < m_Players.size(); ++i)
 			{
@@ -521,7 +521,7 @@ public:
 		return true;
 	}
 
-	void UpdateGameState(const ScriptInterface::StructuredClone& gameState)
+	void UpdateGameState(const Script::StructuredClone& gameState)
 	{
 		ENSURE(m_CommandsComputed);
 		m_GameState = gameState;
@@ -547,7 +547,7 @@ public:
 
 		ScriptRequest rq(m_ScriptInterface);
 		if (dimensionChange || justDeserialized)
-			ScriptInterface::ToJSVal(rq, &m_PassabilityMapVal, m_PassabilityMap);
+			Script::ToJSVal(rq, &m_PassabilityMapVal, m_PassabilityMap);
 		else
 		{
 			// Avoid a useless memory reallocation followed by a garbage collection.
@@ -575,7 +575,7 @@ public:
 
 		ScriptRequest rq(m_ScriptInterface);
 		if (dimensionChange)
-			ScriptInterface::ToJSVal(rq, &m_TerritoryMapVal, m_TerritoryMap);
+			Script::ToJSVal(rq, &m_TerritoryMapVal, m_TerritoryMap);
 		else
 		{
 			// Avoid a useless memory reallocation followed by a garbage collection.
@@ -627,13 +627,13 @@ public:
 
 		m_HasLoadedEntityTemplates = true;
 
-		ScriptInterface::CreateObject(rq, &m_EntityTemplates);
+		Script::CreateObject(rq, &m_EntityTemplates);
 
 		JS::RootedValue val(rq.cx);
 		for (size_t i = 0; i < templates.size(); ++i)
 		{
 			templates[i].second->ToJSVal(rq, false, &val);
-			m_ScriptInterface->SetProperty(m_EntityTemplates, templates[i].first.c_str(), val, true);
+			Script::SetProperty(rq, m_EntityTemplates, templates[i].first.c_str(), val, true);
 		}
 	}
 
@@ -681,7 +681,7 @@ public:
 			for (size_t j = 0; j < m_Players[i]->m_Commands.size(); ++j)
 			{
 				JS::RootedValue val(rq.cx);
-				m_ScriptInterface->ReadStructuredClone(m_Players[i]->m_Commands[j], &val);
+				Script::ReadStructuredClone(rq, m_Players[i]->m_Commands[j], &val);
 				serializer.ScriptVal("command", &val);
 			}
 
@@ -746,7 +746,7 @@ public:
 			{
 				JS::RootedValue val(rq.cx);
 				deserializer.ScriptVal("command", &val);
-				m_Players.back()->m_Commands.push_back(m_ScriptInterface->WriteStructuredClone(val));
+				m_Players.back()->m_Commands.push_back(Script::WriteStructuredClone(rq, val));
 			}
 
 			deserializer.ScriptObjectAssign("data", m_Players.back()->m_Obj);
@@ -786,7 +786,7 @@ private:
 		if (m_PlayerMetadata.find(path) == m_PlayerMetadata.end())
 		{
 			// Load and cache the AI player metadata
-			m_ScriptInterface->ReadJSONFile(path, out);
+			Script::ReadJSONFile(ScriptRequest(m_ScriptInterface), path, out);
 			m_PlayerMetadata[path] = JS::Heap<JS::Value>(out);
 			return;
 		}
@@ -800,13 +800,13 @@ private:
 		JS::RootedValue state(rq.cx);
 		{
 			PROFILE3("AI compute read state");
-			m_ScriptInterface->ReadStructuredClone(m_GameState, &state);
-			m_ScriptInterface->SetProperty(state, "passabilityMap", m_PassabilityMapVal, true);
-			m_ScriptInterface->SetProperty(state, "territoryMap", m_TerritoryMapVal, true);
+			Script::ReadStructuredClone(rq, m_GameState, &state);
+			Script::SetProperty(rq, state, "passabilityMap", m_PassabilityMapVal, true);
+			Script::SetProperty(rq, state, "territoryMap", m_TerritoryMapVal, true);
 		}
 
 		// It would be nice to do
-		//   m_ScriptInterface->FreezeObject(state.get(), true);
+		//   Script::FreezeObject(rq, state.get(), true);
 		// to prevent AI scripts accidentally modifying the state and
 		// affecting other AI scripts they share it with. But the performance
 		// cost is far too high, so we won't do that.
@@ -815,7 +815,7 @@ private:
 		if (m_HasSharedComponent)
 		{
 			PROFILE3("AI run shared component");
-			m_ScriptInterface->CallFunctionVoid(m_SharedAIObj, "onUpdate", state);
+			ScriptFunction::CallVoid(rq, m_SharedAIObj, "onUpdate", state);
 		}
 
 		for (size_t i = 0; i < m_Players.size(); ++i)
@@ -831,19 +831,15 @@ private:
 		}
 	}
 
-	// Take care to keep this declaration before heap rooted members. Destructors of heap rooted
-	// members have to be called before the context destructor.
-	shared_ptr<ScriptContext> m_ScriptContext;
-
-	shared_ptr<ScriptInterface> m_ScriptInterface;
+	std::shared_ptr<ScriptInterface> m_ScriptInterface;
 	boost::rand48 m_RNG;
 	u32 m_TurnNum;
 
 	JS::PersistentRootedValue m_EntityTemplates;
 	bool m_HasLoadedEntityTemplates;
 
-	std::map<VfsPath, JS::Heap<JS::Value> > m_PlayerMetadata;
-	std::vector<shared_ptr<CAIPlayer> > m_Players; // use shared_ptr just to avoid copying
+	std::map<VfsPath, JS::Heap<JS::Value>> m_PlayerMetadata;
+	std::vector<std::shared_ptr<CAIPlayer>> m_Players; // use shared_ptr just to avoid copying
 
 	bool m_HasSharedComponent;
 	JS::PersistentRootedValue m_SharedAIObj;
@@ -851,7 +847,7 @@ private:
 
 	std::set<std::wstring> m_LoadedModules;
 
-	ScriptInterface::StructuredClone m_GameState;
+	Script::StructuredClone m_GameState;
 	Grid<NavcellData> m_PassabilityMap;
 	JS::PersistentRootedValue m_PassabilityMapVal;
 	Grid<u8> m_TerritoryMap;
@@ -871,7 +867,7 @@ private:
 /**
  * Implementation of ICmpAIManager.
  */
-class CCmpAIManager : public ICmpAIManager
+class CCmpAIManager final : public ICmpAIManager
 {
 public:
 	static void ClassInit(CComponentManager& UNUSED(componentManager))
@@ -885,18 +881,20 @@ public:
 		return "<a:component type='system'/><empty/>";
 	}
 
-	virtual void Init(const CParamNode& UNUSED(paramNode))
+	void Init(const CParamNode& UNUSED(paramNode)) override
 	{
+		m_Worker.Init(GetSimContext().GetScriptInterface());
+
 		m_TerritoriesDirtyID = 0;
 		m_TerritoriesDirtyBlinkingID = 0;
 		m_JustDeserialized = false;
 	}
 
-	virtual void Deinit()
+	void Deinit() override
 	{
 	}
 
-	virtual void Serialize(ISerializer& serialize)
+	void Serialize(ISerializer& serialize) override
 	{
 		serialize.NumberU32_Unbounded("num ais", m_Worker.getPlayerSize());
 
@@ -908,7 +906,7 @@ public:
 		m_Worker.Serialize(serialize.GetStream(), serialize.IsDebug());
 	}
 
-	virtual void Deserialize(const CParamNode& paramNode, IDeserializer& deserialize)
+	void Deserialize(const CParamNode& paramNode, IDeserializer& deserialize) override
 	{
 		Init(paramNode);
 
@@ -922,7 +920,7 @@ public:
 		m_JustDeserialized = true;
 	}
 
-	virtual void AddPlayer(const std::wstring& id, player_id_t player, u8 difficulty, const std::wstring& behavior)
+	void AddPlayer(const std::wstring& id, player_id_t player, u8 difficulty, const std::wstring& behavior) override
 	{
 		LoadUsedEntityTemplates();
 
@@ -936,17 +934,17 @@ public:
 			cmpRangeManager->SetLosRevealAll(player, true);
 	}
 
-	virtual void SetRNGSeed(u32 seed)
+	void SetRNGSeed(u32 seed) override
 	{
 		m_Worker.SetRNGSeed(seed);
 	}
 
-	virtual void TryLoadSharedComponent()
+	void TryLoadSharedComponent() override
 	{
 		m_Worker.TryLoadSharedComponent();
 	}
 
-	virtual void RunGamestateInit()
+	void RunGamestateInit() override
 	{
 		const ScriptInterface& scriptInterface = GetSimContext().GetScriptInterface();
 		ScriptRequest rq(scriptInterface);
@@ -979,11 +977,11 @@ public:
 		if (cmpPathfinder)
 			cmpPathfinder->GetPassabilityClasses(nonPathfindingPassClassMasks, pathfindingPassClassMasks);
 
-		m_Worker.RunGamestateInit(scriptInterface.WriteStructuredClone(state),
+		m_Worker.RunGamestateInit(Script::WriteStructuredClone(rq, state),
 			*passabilityMap, *territoryMap, nonPathfindingPassClassMasks, pathfindingPassClassMasks);
 	}
 
-	virtual void StartComputation()
+	void StartComputation() override
 	{
 		PROFILE("AI setup");
 
@@ -1005,7 +1003,7 @@ public:
 		LoadPathfinderClasses(state); // add the pathfinding classes to it
 
 		// Update the game state
-		m_Worker.UpdateGameState(scriptInterface.WriteStructuredClone(state));
+		m_Worker.UpdateGameState(Script::WriteStructuredClone(rq, state));
 
 		// Update the pathfinding data
 		CmpPtr<ICmpPathfinder> cmpPathfinder(GetSystemEntity());
@@ -1042,7 +1040,7 @@ public:
 		m_JustDeserialized = false;
 	}
 
-	virtual void PushCommands()
+	void PushCommands() override
 	{
 		std::vector<CAIWorker::SCommandSets> commands;
 		m_Worker.GetCommands(commands);
@@ -1059,7 +1057,7 @@ public:
 		{
 			for (size_t j = 0; j < commands[i].commands.size(); ++j)
 			{
-				scriptInterface.ReadStructuredClone(commands[i].commands[j], &clonedCommandVal);
+				Script::ReadStructuredClone(rq, commands[i].commands[j], &clonedCommandVal);
 				cmpCommandQueue->PushLocalCommand(commands[i].player, clonedCommandVal);
 			}
 		}
@@ -1106,14 +1104,14 @@ private:
 		ScriptRequest rq(scriptInterface);
 
 		JS::RootedValue classesVal(rq.cx);
-		ScriptInterface::CreateObject(rq, &classesVal);
+		Script::CreateObject(rq, &classesVal);
 
 		std::map<std::string, pass_class_t> classes;
 		cmpPathfinder->GetPassabilityClasses(classes);
 		for (std::map<std::string, pass_class_t>::iterator it = classes.begin(); it != classes.end(); ++it)
-			scriptInterface.SetProperty(classesVal, it->first.c_str(), it->second, true);
+			Script::SetProperty(rq, classesVal, it->first.c_str(), it->second, true);
 
-		scriptInterface.SetProperty(state, "passabilityClasses", classesVal, true);
+		Script::SetProperty(rq, state, "passabilityClasses", classesVal, true);
 	}
 
 	CAIWorker m_Worker;
