@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -37,7 +37,6 @@
 #include "ICmpSound.h"
 
 #include "graphics/Decal.h"
-#include "graphics/Frustum.h"
 #include "graphics/Model.h"
 #include "graphics/ObjectBase.h"
 #include "graphics/ObjectEntry.h"
@@ -45,13 +44,14 @@
 #include "graphics/UnitAnimation.h"
 #include "graphics/UnitManager.h"
 #include "maths/BoundingSphere.h"
+#include "maths/Frustum.h"
 #include "maths/Matrix3D.h"
 #include "maths/Vector3D.h"
 #include "ps/CLogger.h"
 #include "ps/GameSetup/Config.h"
 #include "renderer/Scene.h"
 
-class CCmpVisualActor : public ICmpVisual
+class CCmpVisualActor final : public ICmpVisual
 {
 public:
 	static void ClassInit(CComponentManager& componentManager)
@@ -59,7 +59,6 @@ public:
 		componentManager.SubscribeToMessageType(MT_InterpolatedPositionChanged);
 		componentManager.SubscribeToMessageType(MT_OwnershipChanged);
 		componentManager.SubscribeToMessageType(MT_ValueModification);
-		componentManager.SubscribeToMessageType(MT_TerrainChanged);
 		componentManager.SubscribeToMessageType(MT_Create);
 		componentManager.SubscribeToMessageType(MT_Destroy);
 	}
@@ -192,7 +191,7 @@ public:
 			"</element>";
 	}
 
-	virtual void Init(const CParamNode& paramNode)
+	void Init(const CParamNode& paramNode) override
 	{
 		m_Unit = NULL;
 		m_R = m_G = m_B = fixed::FromInt(1);
@@ -203,7 +202,7 @@ public:
 
 		m_IsFoundationActor = paramNode.GetChild("Foundation").IsOk() && paramNode.GetChild("FoundationActor").IsOk();
 
-		m_BaseActorName = paramNode.GetChild(m_IsFoundationActor ? "FoundationActor" : "Actor").ToString();
+		m_BaseActorName = paramNode.GetChild(m_IsFoundationActor ? "FoundationActor" : "Actor").ToWString();
 		ParseActorName(m_BaseActorName);
 
 		m_VisibleInAtlasOnly = paramNode.GetChild("VisibleInAtlasOnly").ToBool();
@@ -221,7 +220,7 @@ public:
 		InitSelectionShapeDescriptor(paramNode);
 	}
 
-	virtual void Deinit()
+	void Deinit() override
 	{
 		if (m_Unit)
 		{
@@ -253,7 +252,7 @@ public:
 		// TODO: store actor variables?
 	}
 
-	virtual void Serialize(ISerializer& serialize)
+	void Serialize(ISerializer& serialize) override
 	{
 		// TODO: store the actor name, if !debug and it differs from the template
 
@@ -265,7 +264,7 @@ public:
 		SerializeCommon(serialize);
 	}
 
-	virtual void Deserialize(const CParamNode& paramNode, IDeserializer& deserialize)
+	void Deserialize(const CParamNode& paramNode, IDeserializer& deserialize) override
 	{
 		Init(paramNode);
 
@@ -289,7 +288,7 @@ public:
 		}
 	}
 
-	virtual void HandleMessage(const CMessage& msg, bool UNUSED(global))
+	void HandleMessage(const CMessage& msg, bool UNUSED(global)) override
 	{
 		switch (msg.GetType())
 		{
@@ -305,15 +304,6 @@ public:
 
 			break;
 		}
-		case MT_TerrainChanged:
-		{
-			if (!m_Unit)
-				break;
-
-			const CMessageTerrainChanged& msgData = static_cast<const CMessageTerrainChanged&> (msg);
-			m_Unit->GetModel().SetTerrainDirty(msgData.i0, msgData.j0, msgData.i1, msgData.j1);
-			break;
-		}
 		case MT_ValueModification:
 		{
 			// Mirages don't respond to technology modifications.
@@ -359,47 +349,40 @@ public:
 		}
 	}
 
-	virtual CBoundingBoxAligned GetBounds() const
+	CBoundingBoxAligned GetBounds() const override
 	{
 		if (!m_Unit)
 			return CBoundingBoxAligned::EMPTY;
 		return m_Unit->GetModel().GetWorldBounds();
 	}
 
-	virtual CUnit* GetUnit()
+	CUnit* GetUnit() override
 	{
 		return m_Unit;
 	}
 
-	virtual CBoundingBoxOriented GetSelectionBox() const
+	CBoundingBoxOriented GetSelectionBox() const override
 	{
 		if (!m_Unit)
 			return CBoundingBoxOriented::EMPTY;
 		return m_Unit->GetModel().GetSelectionBox();
 	}
 
-	virtual CVector3D GetPosition() const
+	CVector3D GetPosition() const override
 	{
 		if (!m_Unit)
 			return CVector3D(0, 0, 0);
 		return m_Unit->GetModel().GetTransform().GetTranslation();
 	}
 
-	virtual std::wstring GetActorShortName() const
-	{
-		if (!m_Unit)
-			return L"";
-		return m_Unit->GetObject().m_Base->m_ShortName;
-	}
-
-	virtual std::wstring GetProjectileActor() const
+	std::wstring GetProjectileActor() const override
 	{
 		if (!m_Unit)
 			return L"";
 		return m_Unit->GetObject().m_ProjectileModelName;
 	}
 
-	virtual CFixedVector3D GetProjectileLaunchPoint() const
+	CFixedVector3D GetProjectileLaunchPoint() const override
 	{
 		if (!m_Unit)
 			return CFixedVector3D();
@@ -428,7 +411,7 @@ public:
 		return CFixedVector3D();
 	}
 
-	virtual void SetVariant(const CStr& key, const CStr& selection)
+	void SetVariant(const CStr& key, const CStr& selection) override
 	{
 		if (m_VariantSelections[key] == selection)
 			return;
@@ -443,12 +426,12 @@ public:
 		}
 	}
 
-	virtual std::string GetAnimationName() const
+	std::string GetAnimationName() const override
 	{
 		return m_AnimName;
 	}
 
-	virtual void SelectAnimation(const std::string& name, bool once = false, fixed speed = fixed::FromInt(1))
+	void SelectAnimation(const std::string& name, bool once = false, fixed speed = fixed::FromInt(1)) override
 	{
 		m_AnimName = name;
 		m_AnimOnce = once;
@@ -470,15 +453,17 @@ public:
 		m_Unit->GetAnimation()->SetAnimationState(m_AnimName, m_AnimOnce, m_AnimSpeed.ToFloat(), m_AnimDesync.ToFloat(), m_SoundGroup.c_str());	
 	}
 
-	virtual void SelectMovementAnimation(const std::string& name, fixed speed)
+	void SelectMovementAnimation(const std::string& name, fixed speed) override
 	{
 		ENSURE(name == "idle" || name == "walk" || name == "run");
 		if (m_AnimName != "idle" && m_AnimName != "walk" && m_AnimName != "run")
 			return;
+		if (m_AnimName == name && speed == m_AnimSpeed)
+			return;
 		SelectAnimation(name, false, speed);
 	}
 
-	virtual void SetAnimationSyncRepeat(fixed repeattime)
+	void SetAnimationSyncRepeat(fixed repeattime) override
 	{
 		m_AnimSyncRepeatTime = repeattime;
 
@@ -486,7 +471,7 @@ public:
 			m_Unit->GetAnimation()->SetAnimationSyncRepeat(m_AnimSyncRepeatTime.ToFloat());
 	}
 
-	virtual void SetAnimationSyncOffset(fixed actiontime)
+	void SetAnimationSyncOffset(fixed actiontime) override
 	{
 		m_AnimSyncOffsetTime = actiontime;
 
@@ -494,7 +479,7 @@ public:
 			m_Unit->GetAnimation()->SetAnimationSyncOffset(m_AnimSyncOffsetTime.ToFloat());
 	}
 
-	virtual void SetShadingColor(fixed r, fixed g, fixed b, fixed a)
+	void SetShadingColor(fixed r, fixed g, fixed b, fixed a) override
 	{
 		m_R = r;
 		m_G = g;
@@ -508,18 +493,18 @@ public:
 		}
 	}
 
-	virtual void SetVariable(const std::string& name, float value)
+	void SetVariable(const std::string& name, float value) override
 	{
 		if (m_Unit)
 			m_Unit->GetModel().SetEntityVariable(name, value);
 	}
 
-	virtual u32 GetActorSeed() const
+	u32 GetActorSeed() const override
 	{
 		return m_Seed;
 	}
 
-	virtual void SetActorSeed(u32 seed)
+	void SetActorSeed(u32 seed) override
 	{
 		if (seed == m_Seed)
 			return;
@@ -528,7 +513,7 @@ public:
 		ReloadActor();
 	}
 
-	virtual void RecomputeActorName()
+	void RecomputeActorName() override
 	{
 		CmpPtr<ICmpValueModificationManager> cmpValueModificationManager(GetSystemEntity());
 		std::wstring newActorName;
@@ -544,17 +529,17 @@ public:
 		}
 	}
 
-	virtual bool HasConstructionPreview() const
+	bool HasConstructionPreview() const override
 	{
 		return m_ConstructionPreview;
 	}
 
-	virtual void Hotload(const VfsPath& name)
+	void Hotload(const VfsPath& name) override
 	{
 		if (!m_Unit)
 			return;
 
-		if (name != m_ActorName)
+		if (!name.empty() && name != m_ActorName)
 			return;
 
 		ReloadActor();
@@ -603,11 +588,10 @@ void CCmpVisualActor::InitModel()
 	if (!GetSimContext().HasUnitManager())
 		return;
 
-	std::set<CStr> selections;
 	std::wstring actorName = m_ActorName;
 	if (actorName.find(L".xml") == std::wstring::npos)
 		actorName += L".xml";
-	m_Unit = GetSimContext().GetUnitManager().CreateUnit(actorName, GetActorSeed(), selections);
+	m_Unit = GetSimContext().GetUnitManager().CreateUnit(actorName, GetActorSeed());
 	if (!m_Unit)
 		return;
 
@@ -758,6 +742,18 @@ void CCmpVisualActor::ReloadActor()
 
 	InitModel();
 
+	if (!m_Unit)
+	{
+		if (m_ModelTag.valid())
+		{
+			CmpPtr<ICmpUnitRenderer> cmpModelRenderer(GetSystemEntity());
+			if (cmpModelRenderer)
+				cmpModelRenderer->RemoveUnit(m_ModelTag);
+			m_ModelTag = ICmpUnitRenderer::tag_t{};
+		}
+		return;
+	}
+
 	ReloadUnitAnimation();
 
 	m_Unit->GetModel().SetShadingColor(shading);
