@@ -1,4 +1,4 @@
-/* Copyright (C) 2020 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -343,8 +343,6 @@ enum
 	ID_BigScreenshot,
 	ID_JavaScript,
 	ID_CameraReset,
-	ID_RenderPathShaderARB,
-	ID_RenderPathShaderGLSL,
 	ID_DumpState,
 	ID_DumpBinaryState,
 
@@ -379,8 +377,6 @@ BEGIN_EVENT_TABLE(ScenarioEditor, wxFrame)
 	EVT_MENU(ID_CameraReset, ScenarioEditor::OnCameraReset)
 	EVT_MENU(ID_DumpState, ScenarioEditor::OnDumpState)
 	EVT_MENU(ID_DumpBinaryState, ScenarioEditor::OnDumpState)
-	EVT_MENU(ID_RenderPathShaderARB, ScenarioEditor::OnRenderPath)
-	EVT_MENU(ID_RenderPathShaderGLSL, ScenarioEditor::OnRenderPath)
 
 	EVT_MENU(ID_Manual, ScenarioEditor::OnHelp)
 	EVT_MENU(ID_ReportBug, ScenarioEditor::OnHelp)
@@ -499,11 +495,6 @@ ScenarioEditor::ScenarioEditor(wxWindow* parent)
 		menuMisc->AppendSubMenu(menuSS, _("Si&mulation state"));
 		menuSS->Append(ID_DumpState, _("&Dump to disk"));
 		menuSS->Append(ID_DumpBinaryState, _("Dump &binary to disk"));
-
-		wxMenu *menuRP = new wxMenu;
-		menuMisc->AppendSubMenu(menuRP, _("Render &path"));
-		menuRP->Append(ID_RenderPathShaderARB, _("Shader &ARB"));
-		menuRP->Append(ID_RenderPathShaderGLSL, _("&Shader GLSL (default)"));
 	}
 
 	wxMenu *menuHelp = new wxMenu;
@@ -589,10 +580,15 @@ ScenarioEditor::ScenarioEditor(wxWindow* parent)
 
 	// Send setup messages to game engine:
 
+	// We need to set the window HWND for the error dialog instead of iteraing
+	// through all open windows on Windows.
+	POST_MESSAGE(InitAppWindow, (GetHandle()));
+
 	POST_MESSAGE(InitSDL, ());
 
 	POST_MESSAGE(SetCanvas, (static_cast<wxGLCanvas*>(canvas),
-		canvas->GetClientSize().GetWidth(), canvas->GetClientSize().GetHeight()));
+		canvas->GetClientSize().GetWidth() * canvas->GetContentScaleFactor(),
+		canvas->GetClientSize().GetHeight() * canvas->GetContentScaleFactor()));
 
 	POST_MESSAGE(InitGraphics, ());
 
@@ -667,6 +663,10 @@ void ScenarioEditor::OnClose(wxCloseEvent& event)
 
 	m_FileHistory.SaveToSubDir(*wxConfigBase::Get());
 
+	// We notify all clients that might interact with the game after its
+	// shutdown to prevent accessing invalid state.
+	m_SectionLayout.OnShutdown();
+
 	POST_MESSAGE(Shutdown, ());
 
 	qExit().Post();
@@ -748,7 +748,7 @@ bool ScenarioEditor::OpenFile(const wxString& name, const wxString& filename)
 	wxBusyInfo busy(_("Loading ") + name);
 	wxBusyCursor busyc;
 
-	AtlasMessage::qVFSFileExists qry(filename.wc_str());
+	AtlasMessage::qVFSFileExists qry(filename.ToStdWstring());
 	qry.Post();
 	if (!qry.exists)
 		return false;
@@ -758,7 +758,7 @@ bool ScenarioEditor::OpenFile(const wxString& name, const wxString& filename)
 	m_ToolManager.SetCurrentTool(_T(""));
 	// TODO: clear the undo buffer, etc
 
-	std::wstring map(filename.wc_str());
+	std::wstring map(filename.ToStdWstring());
 	POST_MESSAGE(LoadMap, (map));
 
 	SetOpenFilename(name);
@@ -811,7 +811,7 @@ void ScenarioEditor::OnImportHeightmap(wxCommandEvent& WXUNUSED(event))
 
 	OpenFile(_T(""), _T("maps/scenarios/_default.xml"));
 
-	std::wstring image(dlg.GetPath().wc_str());
+	std::wstring image(dlg.GetPath().ToStdWstring());
 	POST_MESSAGE(ImportHeightmap, (image));
 
 	// TODO: Make this a non-undoable command
@@ -855,7 +855,7 @@ void ScenarioEditor::OnSave(wxCommandEvent& event)
 		// the preview units.)
 		m_ToolManager.SetCurrentTool(_T(""));
 
-		std::wstring map(m_OpenFilename.wc_str());
+		std::wstring map(m_OpenFilename.ToStdWstring());
 		POST_MESSAGE(SaveMap, (map));
 
 		// Wait for it to finish saving
@@ -877,7 +877,7 @@ void ScenarioEditor::OnSaveAs(wxCommandEvent& WXUNUSED(event))
 
 		m_ToolManager.SetCurrentTool(_T(""));
 
-		std::wstring map(filePath.wc_str());
+		std::wstring map(filePath.ToStdWstring());
 		POST_MESSAGE(SaveMap, (map));
 
 		SetOpenFilename(filePath);
@@ -932,10 +932,10 @@ void ScenarioEditor::OnScreenshot(wxCommandEvent& event)
 	switch (event.GetId())
 	{
 	case ID_BigScreenshot:
-		POST_MESSAGE(Screenshot, (true, 10));
+		POST_MESSAGE(Screenshot, (true));
 		break;
 	case ID_Screenshot:
-		POST_MESSAGE(Screenshot, (false, 0));
+		POST_MESSAGE(Screenshot, (false));
 		break;
 	}
 }
@@ -953,21 +953,6 @@ void ScenarioEditor::OnCameraReset(wxCommandEvent& WXUNUSED(event))
 	POST_MESSAGE(CameraReset, ());
 }
 
-void ScenarioEditor::OnRenderPath(wxCommandEvent& event)
-{
-	switch (event.GetId())
-	{
-	case ID_RenderPathShaderARB:
-		POST_MESSAGE(SetViewParamS, (eRenderView::GAME, L"renderpath", L"shader"));
-		POST_MESSAGE(SetViewParamB, (eRenderView::GAME, L"preferGLSL", false));
-		break;
-	case ID_RenderPathShaderGLSL:
-		POST_MESSAGE(SetViewParamS, (eRenderView::GAME, L"renderpath", L"shader"));
-		POST_MESSAGE(SetViewParamB, (eRenderView::GAME, L"preferGLSL", true));
-		break;
-	}
-}
-
 void ScenarioEditor::OnDumpState(wxCommandEvent& event)
 {
 	wxDateTime time = wxDateTime::Now();
