@@ -1,4 +1,4 @@
-/* Copyright (C) 2019 Wildfire Games.
+/* Copyright (C) 2022 Wildfire Games.
  * This file is part of 0 A.D.
  *
  * 0 A.D. is free software: you can redistribute it and/or modify
@@ -583,12 +583,11 @@ void PlayerSettingsControl::CreateWidgets()
 	wxArrayString civCodes;
 	AtlasMessage::qGetCivData qryCiv;
 	qryCiv.Post();
-	std::vector<std::string> civData = *qryCiv.data;
-	for (size_t i = 0; i < civData.size(); ++i)
+	std::vector<std::vector<std::wstring>> civData = *qryCiv.data;
+	for (const std::vector<std::wstring>& civ : civData)
 	{
-		AtObj civ = AtlasObject::LoadFromJSON(civData[i]);
-		civNames.Add(wxString::FromUTF8(civ["Name"]));
-		civCodes.Add(wxString::FromUTF8(civ["Code"]));
+		civCodes.Add(civ[0]);
+		civNames.Add(civ[1]);
 	}
 
 	// Load AI data
@@ -858,10 +857,10 @@ AtObj PlayerSettingsControl::UpdateSettingsObject()
 		if (choice->IsEnabled() && choice->GetSelection() >= 0)
 		{
 			wxStringClientData* str = dynamic_cast<wxStringClientData*>(choice->GetClientObject(choice->GetSelection()));
-			player.set("Civ", str->GetData());
+			player.set("Civ", str->GetData().utf8_str());
 		}
 		else
-			player.set("Civ", playerDefs["Civ"]);
+			player.unset("Civ");
 
 		// color
 		if (controls.color->IsEnabled())
@@ -876,17 +875,14 @@ AtObj PlayerSettingsControl::UpdateSettingsObject()
 
 		// player type
 		choice = controls.ai;
-		if (choice->IsEnabled())
+		if (choice->IsEnabled() && choice->GetSelection() > 0)
 		{
-			if (choice->GetSelection() > 0)
-			{
-				// ai - get id
-				wxStringClientData* str = dynamic_cast<wxStringClientData*>(choice->GetClientObject(choice->GetSelection()));
-				player.set("AI", str->GetData());
-			}
-			else // human
-				player.set("AI", _(""));
+			// ai - get id
+			wxStringClientData* str = dynamic_cast<wxStringClientData*>(choice->GetClientObject(choice->GetSelection()));
+			player.set("AI", str->GetData().utf8_str());
 		}
+		else // human
+			player.unset("AI");
 
 		// resources
 		AtObj resObj;
