@@ -4,8 +4,8 @@ from typing import List
 
 from i18n_helper.catalog import Catalog
 
-def getCatalogs(inputFilePath, filters = None) -> List[Catalog]:
-    """Returns a list of "real" catalogs (.po) in the fiven folder."""
+def getCatalogs(inputFilePath, filters : List[str] = None) -> List[Catalog]:
+    """Returns a list of "real" catalogs (.po) in the given folder."""
     existingTranslationCatalogs = []
     l10nFolderPath = os.path.dirname(inputFilePath)
     inputFileName = os.path.basename(inputFilePath)
